import logging
import re
from collections import ChainMap, UserDict
from uuid import uuid4

import msgpack
import msgpack_numpy
import redis


msgpack_numpy.patch()

redis_dict_log = logging.getLogger("nslsii.md_dict.RunEngineRedisDict")


class RunEngineRedisDict(UserDict):
    """
    A class for storing RunEngine metadata added to RE.md on a Redis server.

    This class has two strong ideas about the metadata it manages:
      1. Some key-values are considered "global" or "facility-wide". These
         are in use at all NSLS-II bluesky beamlines and include
         proposal_id, data_session, cycle, SAF, and scan_id. The "global"
         key-values are stored as Redis key-values. Redis does not support
         numeric types, so the RunEngineRedisDict also keeps track of the
         types of the "global" key-values. The intention is that this
         metadata is accessible by any Redis client.
      2. Non-global, or beamline-specific, metadata is stored in Redis as
         a msgpack-ed blob. This means data type conversion between Redis
         and Python is handled by msgpack, including numpy arrays.
         The drawback is that this "local" metadata key-values are not
         directly readable or writeable by Redis clients.
    """

    PACKED_RUNENGINE_METADATA_KEY = "runengine-metadata-blob"

    def __init__(
        self,
        host="localhost",
        port=6379,
        db=0,
        re_md_channel_name="runengine-metadata",
        global_keys=None,
        global_values_types=None,
    ):
        # send no initial data to UserDict.__init__
        # since we will replace UserDict.data entirely
        # with a ChainMap
        super().__init__()
        self._host = host
        self._port = port
        self._db = db
        self._re_md_channel_name = re_md_channel_name
        self._uuid = str(uuid4())

        redis_dict_log.info("connecting to Redis at %s:%s", self._host, self._port)
        # global metadata will be stored as Redis key-value pairs
        # tell the global metadata Redis client to do bytes-to-str conversion
        self._redis_global_client = redis.Redis(
            host=host, port=port, db=db, decode_responses=True
        )
        # ping() will raise redis.exceptions.ConnectionError on failure
        self._redis_global_client.ping()

        # local metadata will be msgpack-ed, so decoding
        # will be handled by msgpack.unpackb()
        # tell the local metadata Redis client to do NO bytes-to-str conversion
        self._redis_local_client = redis.Redis(
            host=host, port=port, db=db, decode_responses=False
        )
        self._redis_local_client.ping()

        if global_keys is None:
            # "global" key-value pairs are
            # present at all NSLS-II beamlines
            self._global_keys = (
                "proposal_id",
                "data_session",
                "cycle",
                "SAF",
                "scan_id",
            )
        else:
            self._global_keys = global_keys

        if global_values_types is None:
            # remember numeric types for global metadata
            # global metadata keys not specified here will default to str
            self._global_values_types = {"scan_id": int}
        else:
            self._global_values_types = global_values_types

        # is local metadata already in redis?
        packed_local_md = self._redis_local_client.get(
            self.PACKED_RUNENGINE_METADATA_KEY
        )
        if packed_local_md is None:
            redis_dict_log.info("no local metadata found in Redis")
            self._local_md = dict()
            self._set_local_metadata_on_server()
        else:
            redis_dict_log.info("unpacking local metadata from Redis")
            self._local_md = self._get_local_metadata_from_server()
            redis_dict_log.debug("unpacked local metadata:\n%s", self._local_md)

        self._global_md = dict()
        for global_key in self._global_keys:
            global_value = self._redis_global_client.get(global_key)
            if global_value is None:
                # if a global key does not exist on the Redis server
                # then it will not exist in the RunEngineRedisDict
                redis_dict_log.info("no value yet for global key %s", global_key)
            else:
                if global_key in self._global_values_types:
                    global_value = self._global_values_types[global_key](global_value)
                self._global_md[global_key] = global_value
        redis_dict_log.info("global metadata: %s", self._global_md)

        # when self._local_md has to be replaced with the metadata
        # blob in Redis we must be careful to replace the first
        # dict in the ChainMap's list of mappings
        self.data = ChainMap(self._local_md, self._global_md)

        # Redis documentation says do not issue commands from
        # a client that has been used to subscribe to a channel
        # so create a client just for subscribing
        self._redis_pubsub_client = redis.Redis(host=host, port=port, db=db)
        self._redis_pubsub_client.ping()
        self._redis_pubsub = self._redis_pubsub_client.pubsub(
            ignore_subscribe_messages=True
        )

        # register self._handle_update_message to handle Redis messages
        # this is how the RunEngineMetadataDict knows a key-value
        # has been modified on the Redis server, and therefore
        # self._local_md must be updated from the server
        self._redis_pubsub.subscribe(
            **{self._re_md_channel_name: self._handle_update_message}
        )
        # start a thread to pass messages to _update_on_message
        self._update_on_message_thread = self._redis_pubsub.run_in_thread(
            sleep_time=0.01, daemon=True
        )

    def __setitem__(self, key, value):
        if key in self._global_keys:
            # can't rely on self._global_md for this check because
            # if global metadata is not in Redis it is not added to self._global_md
            redis_dict_log.debug("setting global metadata %s:%s", key, value)
            # global metadata may be constrained to be of a certain type
            # check that value does not violate the type expected for key
            expected_value_type = self._global_values_types.get(key, str)
            if isinstance(value, expected_value_type):
                # everything is good
                pass
            else:
                raise ValueError(
                    f"expected value for key '{key}' to have type '{expected_value_type}'"
                    f"but '{value}' has type '{type(value)}'"
                )
            # update the global key-value pair explicitly in self._global_md
            # because it can not be updated through the self.data ChainMap
            # since self._global_md is not the first dictionary in that ChainMap
            self._global_md[key] = value
            # update the global key-value pair on the Redis server
            self._redis_global_client.set(name=key, value=value)
        else:
            redis_dict_log.debug("setting local metadata %s:%s", key, value)
            # update the key-value pair in the ChainMap
            super().__setitem__(key, value)
            # update the local key-value pair on the Redis server
            self._redis_local_client.set(
                name=self.PACKED_RUNENGINE_METADATA_KEY,
                value=self._pack(self._local_md),
            )

        # tell subscribers a key-value has changed
        redis_dict_log.debug("publishing update %s:%s", key, value)
        self._publish_metadata_update_message(key)

    def __delitem__(self, key):
        if key in self._global_keys:
            raise KeyError(f"deleting key {key} is not allowed")
        else:
            del self._local_md[key]
            self._set_local_metadata_on_server()

        # tell everyone a (local) key-value has been changed
        self._publish_metadata_update_message(key)

    def _publish_metadata_update_message(self, key):
        """
        Publish a message that includes the updated key and
        the identifying UUID for this RunEngineRedisDict.
        The UUID in the message will allow this RunEngineRedisDict
        to ignore updates that come from itself.
        """
        self._redis_pubsub_client.publish(
            channel=self._re_md_channel_name, message=f"{key}:{self._uuid}"
        )

    def _get_local_metadata_from_server(self):
        return self._unpack(
            self._redis_local_client.get(name=self.PACKED_RUNENGINE_METADATA_KEY)
        )

    def _set_local_metadata_on_server(self):
        self._redis_local_client.set(
            self.PACKED_RUNENGINE_METADATA_KEY, self._pack(self._local_md)
        )

    _message_data_pattern = re.compile(r"^(?P<key>.+):(?P<uuid>.+)$")

    @classmethod
    def _parse_message_data(klass, message):
        """
        message["data"] should look like this
            b"abc:39f1f7fa-aeef-4d83-a802-c1c7f5ff5cb8"
        Splitting the message on ":" gives the updated key
        ("abc" in this example) and the UUID of the RunEngineRedisDict
        that made the update. The UUID is used to determine if
        the update message came from this RunEngineRedisDict, in
        which case it is not necessary to update the local metadata
        from the Redis server.
        """
        decoded_message_data = message["data"].decode()
        message_data_match = klass._message_data_pattern.match(decoded_message_data)

        if message_data_match is None:
            raise ValueError(
                f"message[data]=`{decoded_message_data}` could not be parsed"
            )
        return message_data_match.group("key"), message_data_match.group("uuid")

    def _handle_update_message(self, message):
        redis_dict_log.debug("_update_on_message: %s", message)
        updated_key, publisher_uuid = self._parse_message_data(message)
        if publisher_uuid == self._uuid:
            # this RunEngineRedisDict is the source of this update message,
            # so there is no need to go to the Redis server for the new metadata
            redis_dict_log.debug("update published by me!")
        elif updated_key in self._global_keys:
            redis_dict_log.debug("updated key belongs to global metadata")
            # because the updated_key belongs to "global" metadata
            # we can assume it is not a new or deleted key, so just
            # get the key's value from the Redis database and convert
            # its type if necessary (eg, from string to int)
            updated_value = self._redis_global_client.get(name=updated_key)
            if updated_key in self._global_values_types:
                updated_value = self._global_values_types[updated_key](updated_value)
            self._global_md[updated_key] = updated_value
        else:
            redis_dict_log.debug("updated key belongs to local metadata")
            # the updated key belongs to local metadata
            # it may be a newly added or deleted key, so
            # we have to update the entire local metadata dictionary
            self._local_md = self._get_local_metadata_from_server()
            # update the ChainMap - "local" metadata is always the
            # first element in ChainMap.maps
            self.data.maps[0] = self._local_md

    @staticmethod
    def _pack(obj):
        """Encode as msgpack using numpy-aware encoder."""
        return msgpack.packb(obj)

    @staticmethod
    def _unpack(obj):
        return msgpack.unpackb(obj)



import logging

from collections import namedtuple
from pathlib import Path


def _read_bluesky_kafka_config_file(config_file_path):
    """Read a YAML file of Kafka producer configuration details.

    The file must have three top-level entries as shown:
    ---
        abort_run_on_kafka_exception: true
        bootstrap_servers:
            - kafka1:9092
            - kafka2:9092
        runengine_producer_config:
            acks: 0
            message.timeout.ms: 3000
            compression.codec: snappy

    Parameters
    ----------
    config_file_path: str
        path to the YAML file of Kafka producer configuration details

    Returns
    -------
    dict of configuration details
    """
    import yaml

    # read the Kafka Producer configuration details
    if Path(config_file_path).exists():
        with open(config_file_path) as f:
            bluesky_kafka_config = yaml.safe_load(f)
    else:
        raise FileNotFoundError(config_file_path)

    required_sections = (
        "abort_run_on_kafka_exception",
        "bootstrap_servers",
        # "producer_consumer_security_config",  not required yet
        "runengine_producer_config",
    )
    missing_required_sections = [
        required_section
        for required_section in required_sections
        if required_section not in bluesky_kafka_config
    ]

    if missing_required_sections:
        raise Exception(
            f"Bluesky Kafka configuration file '{config_file_path}' is missing required section(s) `{missing_required_sections}`"
        )

    return bluesky_kafka_config


"""
A namedtuple for holding details of the publisher created by
_subscribe_kafka_publisher.
"""
_SubscribeKafkaPublisherDetails = namedtuple(
    "SubscribeKafkaPublisherDetails",
    {"beamline_topic", "bootstrap_servers", "producer_config", "re_subscribe_token"},
)


def _subscribe_kafka_publisher(
    RE, beamline_name, bootstrap_servers, producer_config, _publisher_factory=None
):
    """
    Subscribe a RunRouter to the specified RE to create Kafka Publishers.
    Each Publisher will publish documents from a single run to the
    Kafka topic "<beamline_name>.bluesky.runengine.documents".

    Parameters
    ----------
    RE: RunEngine
        the RunEngine to which the RunRouter will be subscribed
    beamline_name: str
        beamline start_name, for example "csx", to be used in building the
        Kafka topic to which messages will be published
    bootstrap_servers: str
        Comma-delimited list of Kafka server addresses as a string such as ``'10.0.137.8:9092'``
    producer_config: dict
        dictionary of Kafka Producer configuration settings
    _publisher_factory: callable, optional
        intended only for testing, default is bluesky_kafka.Publisher, optionally specify a callable
        that constructs a Publisher-like object

    Returns
    -------
    topic: str
        the Kafka topic on which bluesky documents will be published
    runrouter_token: int
        subscription token corresponding to the RunRouter subscribed to the RunEngine
        by this function
    """
    from bluesky_kafka import Publisher
    from bluesky_kafka.utils import list_topics
    from event_model import RunRouter

    topic = f"{beamline_name.lower()}.bluesky.runengine.documents"

    if _publisher_factory is None:
        _publisher_factory = Publisher

    def kafka_publisher_factory(start_name, start_doc):
        # create a Kafka Publisher for a single run
        #   in response to a start document

        kafka_publisher = _publisher_factory(
            topic=topic,
            bootstrap_servers=bootstrap_servers,
            key=start_doc["uid"],
            producer_config=producer_config,
            flush_on_stop_doc=True,
        )

        def publish_or_abort_run(name_, doc_):
            """
            Exceptions _should_ interrupt the current run.
            """
            try:
                kafka_publisher(name_, doc_)
            except (BaseException, Exception) as exc_:
                # log the exception and re-raise it to abort the current run
                logger = logging.getLogger("nslsii")
                logger.exception(
                    "an error occurred when %s published\n  start_name: %s\n  doc %s",
                    kafka_publisher,
                    name_,
                    doc_,
                )
                raise exc_

        try:
            # on each start document call list_topics to test if we can connect to a Kafka broker
            cluster_metadata = list_topics(
                bootstrap_servers=bootstrap_servers,
                producer_config=producer_config,
                timeout=5.0,
            )
            logging.getLogger("nslsii").info(
                "connected to Kafka broker(s): %s", cluster_metadata
            )
            return [publish_or_abort_run], []
        except (BaseException, Exception) as exc:
            # log the exception and re-raise it to indicate no connection could be made to a Kafka broker
            nslsii_logger = logging.getLogger("nslsii")
            nslsii_logger.exception("'%s' failed to connect to Kafka", kafka_publisher)
            raise exc

    rr = RunRouter(factories=[kafka_publisher_factory])
    runrouter_token = RE.subscribe(rr)

    # log this only once
    logging.getLogger("nslsii").info(
        "RE will publish documents to Kafka topic '%s'", topic
    )

    subscribe_kafka_publisher_details = _SubscribeKafkaPublisherDetails(
        beamline_topic=topic,
        bootstrap_servers=bootstrap_servers,
        producer_config=producer_config,
        re_subscribe_token=runrouter_token,
    )

    return subscribe_kafka_publisher_details


"""
A namedtuple for holding details of the publisher created by
_subscribe_kafka_queue_thread_publisher.
"""
_SubscribeKafkaQueueThreadPublisherDetails = namedtuple(
    "SubscribeKafkaQueueThreadPublisherDetails",
    {
        "beamline_topic",
        "bootstrap_servers",
        "producer_config",
        "publisher_queue_thread_details",
        "re_subscribe_token",
    },
)


def _subscribe_kafka_queue_thread_publisher(
    RE, beamline_name, bootstrap_servers, producer_config, publisher_queue_timeout=1
):
    """
    Create and start a separate thread to publish bluesky documents as Kafka
    messages on a beamline-specific topic.

    This function performs three tasks:
      1) verify a Kafka broker with the expected beamline-specific topic is available
      2) instantiate a bluesky_kafka.Publisher with the expected beamline-specific topic
      3) delegate connecting the RunEngine and Publisher to _subscribe_kafka_publisher

    Parameters
    ----------
    RE: RunEngine
        the RunEngine to which the RunRouter will be subscribed
    beamline_name: str
        beamline name, for example "csx", to be used in building the
        Kafka topic to which messages will be published
    bootstrap_servers: str
        Comma-delimited list of Kafka server addresses or hostnames and ports as a string
        such as ``'kafka1:9092,kafka2:9092``
    producer_config: dict
        dictionary of Kafka Producer configuration settings

    Returns
    -------
    topic: str
        the Kafka topic on which bluesky documents will be published, for example
        "csx.bluesky.runengine.documents"
    publisher_thread_re_token: int
        RunEngine subscription token corresponding to the function subscribed to the RunEngine
        that places (name, document) tuples on the publisher queue. This token is needed to
        un-subscribe the function from the RunEngine, in case someone ever wants to do that.

    """
    from bluesky_kafka import BlueskyKafkaException
    from bluesky_kafka.tools.queue_thread import build_kafka_publisher_queue_and_thread

    nslsii_logger = logging.getLogger("nslsii")
    beamline_runengine_topic = None
    kafka_publisher_token = None
    publisher_thread_stop_event = None
    kafka_publisher_re_token = None
    publisher_queue_thread_details = None

    try:
        nslsii_logger.info("connecting to Kafka broker(s): '%s'", bootstrap_servers)
        beamline_runengine_topic = (
            f"{beamline_name.lower()}.bluesky.runengine.documents"
        )

        publisher_queue_thread_details = build_kafka_publisher_queue_and_thread(
            topic=beamline_runengine_topic,
            bootstrap_servers=bootstrap_servers,
            producer_config=producer_config,
            publisher_queue_timeout=publisher_queue_timeout,
        )

        publisher_thread_stop_event = (
            publisher_queue_thread_details.publisher_thread_stop_event
        )

        kafka_publisher_re_token = RE.subscribe(
            publisher_queue_thread_details.put_on_publisher_queue
        )

        nslsii_logger.info(
            "RunEngine will publish bluesky documents on Kafka topic '%s'",
            beamline_runengine_topic,
        )

    except BaseException:
        """
        An exception at this point means bluesky documents
        will not be published as Kafka messages.
        The exception will stop here so the run will not be aborted.
        """
        nslsii_logger.exception(
            "RunEngine is not able to publish bluesky documents as Kafka messages on topic '%s'",
            beamline_runengine_topic,
        )

    subscribe_kafka_queue_thread_publisher_details = (
        _SubscribeKafkaQueueThreadPublisherDetails(
            beamline_topic=beamline_runengine_topic,
            bootstrap_servers=bootstrap_servers,
            producer_config=producer_config,
            publisher_queue_thread_details=publisher_queue_thread_details,
            re_subscribe_token=kafka_publisher_re_token,
        )
    )

    return subscribe_kafka_queue_thread_publisher_details



import logging
import re
import time as ttime
from collections import OrderedDict

from databroker.assets.handlers import Xspress3HDF5Handler

from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DynamicDeviceCpt
from ophyd import EpicsSignal, EpicsSignalRO, Signal
from ophyd.areadetector import ADBase
from ophyd.areadetector import EpicsSignalWithRBV as SignalWithRBV
from ophyd.areadetector import Xspress3Detector
from ophyd.areadetector.filestore_mixins import FileStorePluginBase
from ophyd.areadetector.plugins import HDF5Plugin
from ophyd.device import BlueskyInterface, Staged
from ophyd.status import DeviceStatus

from ..detectors.utils import makedirs


logger = logging.getLogger(__name__)


class Xspress3Trigger(BlueskyInterface):
    """
    A basic trigger mixin. Good enough for simple cases.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._acquire_status = None
        self._abs_trigger_count = None

    def stage(self):
        logger.debug("stage")
        self._abs_trigger_count = 0
        self.cam.acquire.subscribe(self._acquire_changed)
        return super().stage()

    def unstage(self):
        logger.debug("unstage")
        super_unstage_result = super().unstage()
        self.cam.acquire.clear_sub(self._acquire_changed)
        self._acquire_status = None
        return super_unstage_result

    def _acquire_changed(self, value=None, old_value=None, **kwargs):
        """Respond to changes in the Xspress3Detector.cam.acquire PV.

        The important behavior of this method is to mark self._acquire_status
        as `finished` when Xspress3Detector.cam.acquire changes from 1 to 0
        (high to low). A RunEngine is waiting for this change.

        Parameters
        ----------
        value: int
            new value of the Xspress3Detector.cam.acquire PV
        old_value: int
            previous value of the Xspress3Detector.cam.acquire PV
        kwargs: dict
            unused

        Returns
        -------
        No return value
        """

        if self._acquire_status is None:
            return
        if (old_value == 1) and (value == 0):
            # Negative-going edge means an acquisition just finished.
            self._acquire_status.set_finished()
            self._acquire_status = None

    def new_acquire_status(self):
        """
        Create, remember, and return a Status object that will be marked
        as `finished` when acquisition is done (see _acquire_changed). The
        intention is that this Status will be used by another object,
        for example a RunEngine.

        Returns
        -------
        DeviceStatus
        """

        self._acquire_status = DeviceStatus(self)
        return self._acquire_status

    def trigger(self):
        logger.debug("trigger")
        if self._staged != Staged.yes:
            raise RuntimeError(
                "tried to trigger Xspress3 with prefix {self.prefix} but it is not staged"
            )

        acquire_status = self.new_acquire_status()
        self.cam.acquire.put(1, wait=False)
        trigger_time = ttime.time()

        # tell the associated file plugins to do something about
        # each channel's data by invoking
        # FileStore.generate_datum(key=channel.name, ...)
        for channel in self.iterate_channels():
            self.dispatch(key=channel.name, timestamp=trigger_time)

        self._abs_trigger_count += 1

        return acquire_status


class Xspress3FileStore(FileStorePluginBase, HDF5Plugin):
    """
    Create resource and datum documents.
    """

    # TODO: these PVs may be obsolete
    num_capture_calc = Cpt(EpicsSignal, "NumCapture_CALC")
    num_capture_calc_disable = Cpt(EpicsSignal, "NumCapture_CALC.DISA")

    filestore_spec = Xspress3HDF5Handler.HANDLER_NAME

    def __init__(
        self,
        basename,
        *,
        stage_sleep_time=0.5,
        # JL: what are mds keys?
        # mds_key_format="{self.parent.cam.name}_ch{channel_number}",
        mds_key_format="{self.parent.name}_channels_channel{channel_number:02}",
        parent,
        **kwargs,
    ):
        """

        Parameters
        ----------
        basename:
        stage_sleep_time: int
            number of seconds to pause at the end of method stage()
        mds_key_format: str
        parent: Xspress3Detector, required
        kwargs:
        """
        super().__init__(basename, parent=parent, **kwargs)

        if not isinstance(parent, Xspress3Detector):
            raise TypeError(
                "parent must be an instance of ophyd.areadetector.Xspress3Detector"
            )

        # establish PV values to be set when this detector is staged
        # the original values will be replaced when it is unstaged
        self.stage_sigs[self.blocking_callbacks] = 1
        self.stage_sigs[self.enable] = 1
        self.stage_sigs[self.compression] = "zlib"
        self.stage_sigs[self.file_template] = "%s%s_%6.6d.h5"

        # allow for a pause when staging this detector
        self.stage_sleep_time = stage_sleep_time

        self._filestore_res = None
        # JL: what are mds_keys?
        self.mds_keys = {
            channel.channel_number: mds_key_format.format(
                self=self, channel_number=channel.channel_number
            )
            for channel in parent.iterate_channels()
        }

    def stop(self, success=False):
        super_stop_return = super().stop(success=success)
        self.capture.put(0)
        return super_stop_return

    def kickoff(self):
        raise NotImplementedError()

    def collect(self):
        raise NotImplementedError()

    def make_filename(self):
        file_name, read_path, write_path = super().make_filename()
        if self.parent.make_directories.get():
            makedirs(write_path)
        return file_name, read_path, write_path

    def stage(self):
        logger.debug("staging Xspress3 '%s'", self.parent.prefix)

        # force the Xspress3 to stop acquiring
        self.parent.cam.acquire.put(0, wait=True)

        total_points_reading = self.parent.total_points.get()
        if total_points_reading < 1:
            raise RuntimeError(f"total_points '{self.parent.total_points}' must be set")
        spectra_per_point_reading = self.parent.spectra_per_point.get()
        total_capture = total_points_reading * spectra_per_point_reading

        # stop previous acquisition
        # JL: this was set above, don't set it again
        # self.stage_sigs[self.parent.cam.acquire] = 0

        # re-order the stage signals and disable the calc record which is
        # interfering with the capture count
        # JL: can we get rid of num_capture_calc_disable?
        # JL: is self.num_capture even in stage_sigs?
        self.stage_sigs.pop(self.num_capture, None)
        self.stage_sigs.pop(self.parent.cam.num_images, None)
        self.stage_sigs[self.num_capture_calc_disable] = 1

        external_trig_reading = self.parent.external_trig.get()
        if external_trig_reading:
            logger.debug(
                "Xspress3 '%s' will be triggered externally", self.parent.prefix
            )
            self.stage_sigs[self.parent.cam.trigger_mode] = "TTL Veto Only"
            self.stage_sigs[self.parent.cam.num_images] = total_capture
        else:
            logger.debug(
                "Xspress3 '%s' will be triggered internally", self.parent.prefix
            )
            self.stage_sigs[self.parent.cam.trigger_mode] = "Internal"
            # JL: why not total_capture as above?
            self.stage_sigs[self.parent.cam.num_images] = spectra_per_point_reading

        self.stage_sigs[self.auto_save] = "No"

        filename, read_path, write_path = self.make_filename()
        logger.debug(
            "xspress3 '%s' read path: '%s' write path: '%s' filename: '%s'",
            self.parent.prefix,
            read_path,
            write_path,
            filename,
        )

        logger.debug("erasing old spectra with '%s'", self.parent.cam.erase)
        self.parent.cam.erase.put(1, wait=True)

        # this must be set after self.parent.cam.num_images because at the EPICS
        # layer there is a helpful link that sets this equal to that (but
        # not the other way)
        # JL: hoping num_capture does not exist on the new IOC
        self.stage_sigs[self.num_capture] = total_capture

        # apply the stage_sigs values
        super_stage_result = super().stage()

        # self._fn comes from FileStorePluginBase
        self._fn = self.file_template.get() % (
            self._fp,
            self.file_name.get(),
            self.file_number.get(),
        )

        if not self.file_path_exists.get():
            raise IOError(f"path '{self.file_path.get()}' does not exist on the IOC")

        logger.debug("inserting the filestore resource: '%s'", self._fn)
        # JL: calling generate_resource() in stage is usual
        self._generate_resource({})
        # JL: the last element of the last element of self._asset_docs_cache
        #     is the "resource"?
        # self._asset_docs_cache looks like this:
        #   [  ("resource", {
        #         "uid": ,
        #         "resource-kwargs" : {key-values here go into the spec'd handler __init__}
        #         "spec": "WHAT KIND OF FILE",
        #         "root": the start of the path,
        #         "resource-path": the rest of the path,
        #         "path-semantics": "posix":,
        #         "run-start-id": }
        #      )
        #      ("datum", {
        #       "datum_id": "resource-uuid/datum-identifier",
        #       "resource_id": "resource"
        #       "datum-kwargs": {...}})
        #      ("datum-page", {}) this is in the future
        #   ]
        self._filestore_res = self._asset_docs_cache[-1][-1]

        # this gets automatically turned off at the end
        self.capture.put(1)

        # Xspress3 needs a bit of time to configure itself...
        # this does not play nice with the event loop :/
        ttime.sleep(self.stage_sleep_time)

        return super_stage_result

    def unstage(self):
        try:
            i = 0
            # this needs a fail-safe, RE will now hang forever here
            # as we eat all SIGINT to ensure that cleanup happens in
            # orderly manner.
            # If we are here this is a sign that we have not configured the xs3
            # correctly and it is expecting to capture more points than it
            # was triggered to take.
            while self.capture.get() == 1:
                i += 1
                if (i % 50) == 0:
                    logger.warning("Still capturing data .... waiting.")
                ttime.sleep(0.1)
                if i > 150:
                    logger.warning("Still capturing data .... giving up.")
                    logger.warning(
                        "Check that the xspress3 is configured to take the right "
                        "number of frames "
                        f"(it is trying to take {self.parent.cam.num_images.get()})"
                    )
                    self.capture.put(0)
                    break

        except KeyboardInterrupt:
            self.capture.put(0)
            logger.warning("Still capturing data .... interrupted.")

        return super().unstage()

    def generate_datum(self, key, timestamp, datum_kwargs):
        """Create datum kwargs for one Xspress3 channel following a trigger().

        Parameters
        ----------
        key: str
            Xspress3 channel 'name', for example "det1_channels_channel01"
        timestamp: float
        datum_kwargs: dict

        Returns
        -------
        no return value
        """

        logger.debug("generate_datum() called with key '%s'", key)
        # find the channel corresponding to `key`
        # and create the corresponding datum_kwargs
        # if we do not find a corresponding channel then we have a problem
        for channel in self.parent.iterate_channels():
            if channel.name == key:
                datum_kwargs.update(
                    {
                        "frame": self.parent._abs_trigger_count,
                        "channel": channel.channel_number,
                    }
                )
                self.mds_keys[channel.channel_number] = key
                super().generate_datum(
                    key=key, timestamp=timestamp, datum_kwargs=datum_kwargs
                )
                # we are done
                return
            else:
                pass

        # we have a problem
        # the `key` parameter did not match any of our channels
        raise ValueError(
            f"failed to find channel with name '{key}' "
            f"on Xspress3 detector with PV prefix '{self.parent.prefix}'"
        )

    # JL: is there any reason to keep this?
    def configure(self, total_points=0, master=None, external_trig=False, **kwargs):
        raise NotImplementedError()

    def describe(self):
        spec_desc = {
            "external": "FILESTORE:",
            "dtype": "array",
            # is there a better value than this for shape?
            "shape": (self.width.get(),),
            "source": "FileStore:",
        }

        # # JL: replace with parent.iterate_channels()
        # desc = OrderedDict()
        # for chan in self.channels:
        #     key = self.mds_keys[chan]
        #     desc[key] = spec_desc

        return {
            self.mds_keys[channel.channel_number]: spec_desc
            for channel in self.parent.iterate_channels()
        }


# start new IOC classes
# are these general areadetector plugins?
# for now they are being used just for xspress3
class Mca(ADBase):
    array_data = Cpt(EpicsSignal, "ArrayData")


class McaSum(ADBase):
    array_data = Cpt(EpicsSignal, "ArrayData")


class McaRoi(ADBase):
    roi_name = Cpt(EpicsSignal, "Name")
    min_x = Cpt(EpicsSignal, "MinX")
    size_x = Cpt(EpicsSignal, "SizeX")
    total_rbv = Cpt(EpicsSignalRO, "Total_RBV")

    use = Cpt(SignalWithRBV, "Use")

    mcaroi_prefix_re = re.compile(
        r"MCA(?P<channel_number>\d+)ROI:(?P<mcaroi_number>\d+):"
    )

    def __init__(self, prefix, *args, **kwargs):
        super().__init__(prefix, *args, **kwargs)
        # peel the 'number' off of the prefix
        # the prefix looks like "MCA1ROI:1:"
        # and we want the 1 at the end
        mcaroi_prefix_match = self.mcaroi_prefix_re.search(prefix)
        if mcaroi_prefix_match is None:
            raise ValueError(
                f"mcaroi prefix '{prefix}' does not match the expected pattern `{self.mcaroi_prefix_re.pattern}`"
            )
        self.mcaroi_number = int(mcaroi_prefix_match.group("mcaroi_number"))

    def configure_mcaroi(self, *, min_x, size_x, roi_name=None, use=True):
        """
        Configure the details of an MCAROI.

        Parameters
        ----------
        min_x: int
            the starting bin? for the roi
        size_x: int
            the width in bins? for the roi
        roi_name: str, optional
        use: bool, defaults to True
        """

        logger.debug(
            "configuring Xspress3 MCAROI '%s': name '%s' min_x '%d' size_x '%d' use '%s'",
            self.prefix,
            roi_name,
            min_x,
            size_x,
            use,
        )
        self.min_x.put(int(min_x))
        self.size_x.put(int(size_x))
        if roi_name:
            self.roi_name.put(roi_name)
        self.use.put(use)

    # remove in favor of configure_mcaroi
    def configure_roi(self, ev_min, ev_size):
        """Configure the MCAROI with min and size eV

        Parameters
        ----------
        ev_min : int
            minimum electron volts for ROI
        ev_size : int
            ROI size (width) in electron volts
        """
        ev_min = int(ev_min)
        ev_size = int(ev_size)

        # assume if this ROI is being configured
        # that it should be read, meaning the
        # "use" PV must be set to 1
        use_roi = 1
        configuration_changed = any(
            [
                self.min_x.get() != ev_min,
                self.size_x.get() != ev_size,
                self.use.get() != use_roi,
            ]
        )

        if configuration_changed:
            logger.debug(
                "Setting up Xspress3 ROI: name=%s ev_min=%s ev_size=%s "
                "use=%s prefix=%s channel=%s",
                self.name,
                ev_min,
                ev_size,
                use_roi,
                self.prefix,
                # self.parent is the ?? class
                # self.parent.parent is the ?? class
                # TODO: I don't like the assumption that self has a parent
                self.parent.parent.channel_num,
            )

            self.min_x.put(ev_min)
            self.size_x.put(ev_size)
            self.use.put(use_roi)
        else:
            # nothing has changed
            pass

    def clear(self):
        """Clear and disable this ROI"""
        # it is enough to just disable the ROI
        # self.min_x.put(0)
        # self.size_x.put(0)
        self.use.put(0)


class Sca(ADBase):
    # includes Dead Time correction, for example
    # sca numbers go from 0 to 10
    clock_ticks = Cpt(EpicsSignalRO, "0:Value_RBV")
    reset_ticks = Cpt(EpicsSignalRO, "1:Value_RBV")
    reset_counts = Cpt(EpicsSignalRO, "2:Value_RBV")
    all_event = Cpt(EpicsSignalRO, "3:Value_RBV")
    all_good = Cpt(EpicsSignalRO, "4:Value_RBV")
    window_1 = Cpt(EpicsSignalRO, "5:Value_RBV")
    window_2 = Cpt(EpicsSignalRO, "6:Value_RBV")
    pileup = Cpt(EpicsSignalRO, "7:Value_RBV")
    event_width = Cpt(EpicsSignalRO, "8:Value_RBV")
    dt_factor = Cpt(EpicsSignalRO, "9:Value_RBV")
    dt_percent = Cpt(EpicsSignalRO, "10:Value_RBV")


def _validate_mcaroi_number(mcaroi_number):
    """
    Raise ValueError if the MCAROI number is
     1. not an integer
     2. outside the allowed interval [1,48]

    Parameters
    ----------
    mcaroi_number: could be anything
        MCAROI number candidate

    """
    if not isinstance(mcaroi_number, int):
        raise ValueError(f"MCAROI number '{mcaroi_number}' is not an integer")
    elif not 1 <= mcaroi_number <= 48:
        raise ValueError(
            f"MCAROI number '{mcaroi_number}' is outside the allowed interval [1,48]"
        )
    else:
        # everything is awesome
        pass


def build_channel_class(channel_number, mcaroi_numbers, channel_parent_classes=None):
    """Build an Xspress3 channel class with the specified channel number and MCAROI numbers.

    MCAROI numbers need not be consecutive.

    The complication of using dynamically generated classes
    is the price for the relative ease of including the channel
    number in MCAROI PVs and the ability to specify the number
    of MCAROIs that will be used rather than defaulting to the
    maximum of 48 per channel.

    Parameters
    ----------
    channel_number: int
        the channel number, 1-16
    mcaroi_numbers: Sequence of int
         sequence of MCAROI numbers, not necessarily consecutive, allowed values are 1-48
    channel_parent_classes: list-like, optional
        sequence of all parent classes for the generated channel class,
        by default the only parent is ophyd.areadetector.ADBase

    Returns
    -------
    a dynamically generated class similar to this:
        class GeneratedXspress3Channel(ADBase):
            channel_num = 2
            sca = Cpt(Sca, ...)
            mca = Cpt(Mca, ...)
            mca_sum = Cpt(McaSum, ...)
            mcarois = DDC(...4 McaRois...)

            def get_mcaroi(self, *, number):
                ...
            def iterate_mcarois(self):
                ...
            def clear_all_rois(self):
                ...

    """
    if channel_parent_classes is None:
        channel_parent_classes = tuple([ADBase])

    _validate_channel_number(channel_number=channel_number)

    # create a tuple in case the mcaroi_numbers parameter can be iterated only once
    mcaroi_numbers = tuple([mcaroi_number for mcaroi_number in mcaroi_numbers])
    for mcaroi_number in mcaroi_numbers:
        _validate_mcaroi_number(mcaroi_number=mcaroi_number)

    mcaroi_name_re = re.compile(r"mcaroi\d{2}")

    # the next six functions will become methods of the generated channel class
    def __repr__(self):
        return f"{self.__class__.__name__}(channel_number={self.channel_number}, mcaroi_numbers={self.mcaroi_numbers})"

    def get_mcaroi_count(self):
        return len(mcaroi_numbers)

    def get_mcaroi(self, *, mcaroi_number):
        _validate_mcaroi_number(mcaroi_number=mcaroi_number)
        try:
            return getattr(self, f"mcaroi{mcaroi_number:02d}")
        except AttributeError as ae:
            raise ValueError(
                f"no MCAROI on channel {self.channel_number} "
                f"with prefix '{self.prefix}' has number {mcaroi_number}"
            ) from ae

    def iterate_mcaroi_attr_names(self):
        for attr_name in self.__dir__():
            if mcaroi_name_re.match(attr_name):
                yield attr_name

    def iterate_mcarois(self):
        """
        Iterate over McaRoi children of the Xspress3Channel.mcarois attribute.

        Yields
        ------
        McaRoi instance
        """
        for mcaroi_name, mcaroi in self._signals.items():
            if mcaroi_name_re.match(mcaroi_name):
                yield mcaroi

    def clear_all_rois(self):
        """Clear all MCAROIs"""
        for mcaroi in self.iterate_mcarois():
            mcaroi.clear()

    channel_fields_and_methods = {
        "__repr__": __repr__,
        "channel_number": channel_number,
        "mcaroi_numbers": tuple(sorted(mcaroi_numbers)),
        "sca": Cpt(Sca, f"C{channel_number}SCA:"),
        "mca": Cpt(Mca, f"MCA{channel_number}:"),
        "mca_sum": Cpt(McaSum, f"MCASUM{channel_number}:"),
        "get_mcaroi_count": get_mcaroi_count,
        "get_mcaroi": get_mcaroi,
        "iterate_mcaroi_attr_names": iterate_mcaroi_attr_names,
        "iterate_mcarois": iterate_mcarois,
        "clear_all_rois": clear_all_rois,
    }

    channel_fields_and_methods.update(
        {
            f"mcaroi{mcaroi_i:02d}": Cpt(
                McaRoi,
                # MCAROI PV suffixes look like "MCA1ROI:2:"
                f"MCA{channel_number}ROI:{mcaroi_i:d}:",
            )
            for mcaroi_i in mcaroi_numbers
        }
    )

    return type(
        "GeneratedXspress3Channel", channel_parent_classes, channel_fields_and_methods
    )


def _validate_channel_number(channel_number):
    """
    Raise ValueError if the channel number is
     1. not an integer
     2. outside the allowed interval [1,16]

    Parameters
    ----------
    channel_number: could be anything, but should be int
        channel number candidate

    """
    if not isinstance(channel_number, int):
        raise ValueError(f"channel number '{channel_number}' is not an integer")
    elif not 1 <= channel_number <= 16:
        raise ValueError(
            f"channel number '{channel_number}' is outside the allowed interval [1,16]"
        )
    else:
        # everything is great
        pass


def build_detector_class(
    channel_numbers,
    mcaroi_numbers,
    detector_parent_classes=None,
    extra_class_members=None,
):
    return build_xspress3_class(
        channel_numbers=channel_numbers,
        mcaroi_numbers=mcaroi_numbers,
        xspress3_parent_classes=detector_parent_classes,
        extra_class_members=extra_class_members,
    )


def build_xspress3_class(
    channel_numbers,
    mcaroi_numbers,
    xspress3_parent_classes=None,
    extra_class_members=None,
):
    """Build an Xspress3 detector class with the specified channel and roi numbers.

    The complication of using dynamically generated detector classes
    is the price for being able to easily specify the exact number of
    channels and MCAROIs per channel in use on the detector.

    Detector classes generated by build_detector_class include these "soft" PVs
    which are not part of the Xspress3 IOC but are used by Xspress3FileStore:
        external_trig
        total_points
        spectra_per_point
        make_directories
        rewindable

    Parameters
    ----------
    channel_numbers: Sequence of int
        sequence of channel numbers, 1-16, for the detector; for example [1, 2, 3, 8]
    mcaroi_numbers: Sequence of int
        sequence of MCAROI numbers, 1-48, for each channel; for example [1, 2, 3, 10]
    xspress3_parent_classes: list-like, optional
        sequence of all parent classes for the generated detector class,
        if specified include *all* necessary parent classes; if not specified
        the default parent is ophyd.areadetector.Xspress3Detector
    extra_class_members: Dict[String, Any]
        a dictionary of extra class members to be passed to the builtin type(...)
        function; see the builtin type function for allowed key-value pairs

    Returns
    -------
    a dynamically generated class similar to the following:
        class GeneratedXspress3Detector(Xspress3Detector, SomeMixinClass, ...):
            external_trig = Cpt(Signal, value=False)
            total_points = Cpt(Signal, value=-1)
            spectra_per_point = Cpt(Signal, value=1)
            make_directories = Cpt(Signal, value=False)
            rewindable = Cpt(Signal, value=False)
            channels = DDC(...4 Xspress3Channels with 3 ROIs each...)

            def get_channel_count(self):
                ....
            def get_channel(self, channel_number):
                ....
            def iterate_channels(self):
                ...
    """
    if xspress3_parent_classes is None:
        xspress3_parent_classes = tuple([Xspress3Detector])

    if extra_class_members is None:
        extra_class_members = dict()

    # in case channel_numbers can be iterated only once, create a tuple
    channel_numbers = tuple([channel_number for channel_number in channel_numbers])

    # in case mcaroi_numbers can be iterated only once, create a tuple
    mcaroi_numbers = tuple([mcaroi_number for mcaroi_number in mcaroi_numbers])

    channel_attr_name_re = re.compile(r"channel\d{2}")

    # the following four functions will become methods of the generated xspress3 class
    def __repr__(self):
        """Return a string representation of this xspress3 class.

        Returns
        -------
        str : text representation of the dynamically generated xspress3 class
        """
        return f"{self.__class__.__name__}(channels=({','.join([str(channel) for channel in self.iterate_channels()])}))"

    def get_channel_count(self):
        """Return the number of channels on this xspress3 class.

        Returns
        -------
        int : count of channels on this xspress3 class
        """
        return len(channel_numbers)

    def get_channel(self, *, channel_number):
        """Return the channel object corresponding to the specified channel number.

        Parameters
        ----------
        channel_number
            integer channel number

        Returns
        -------
        channel : GeneratedXspress3Channel

        Raises
        ------
        ValueError
            when there is no channel with the specified channel number
        """
        _validate_channel_number(channel_number=channel_number)
        try:
            return getattr(self, f"channel{channel_number:02d}")
        except AttributeError as ae:
            raise ValueError(
                f"no channel on detector with prefix '{self.prefix}' "
                f"has number {channel_number}"
            ) from ae

    def iterate_channels(self):
        """Yield the channel objects of this xspress3 class in the order they were specified.

        Yields
        ------
        channel : GeneratedXspress3Channel
        """

        for channel_attr_name in self.__dir__():
            if channel_attr_name_re.match(channel_attr_name):
                yield getattr(self, channel_attr_name)

    xspress3_fields_and_methods = dict(
        **{
            "channel_numbers": tuple(sorted(channel_numbers)),
            "external_trig": Cpt(Signal, value=False, doc="Use external triggering"),
            "total_points": Cpt(
                Signal,
                value=-1,
                doc="The total number of points to acquire overall",
            ),
            "spectra_per_point": Cpt(
                Signal, value=1, doc="Number of spectra per point"
            ),
            "make_directories": Cpt(
                Signal, value=False, doc="Make directories on the Xspress3 side"
            ),
            "rewindable": Cpt(
                Signal,
                value=False,
                doc="Xspress3 cannot safely be rewound in bluesky",
            ),
            "__repr__": __repr__,
            "get_channel_count": get_channel_count,
            "get_channel": get_channel,
            "iterate_channels": iterate_channels,
        },
        **extra_class_members,
    )

    xspress3_fields_and_methods.update(
        {
            f"channel{c:02d}": Cpt(
                build_channel_class(channel_number=c, mcaroi_numbers=mcaroi_numbers),
                # there is no discrete channel prefix
                # for the Xspress3 IOC PVs
                # so specify an empty string here
                "",
            )
            for c in channel_numbers
        }
    )

    return type(
        "GeneratedXspress3Detector",
        xspress3_parent_classes,
        xspress3_fields_and_methods,
    )



from matplotlib import animation
from matplotlib import pyplot as plt
from IPython.display import HTML
from ipywidgets import interact
from tempfile import NamedTemporaryFile
import base64


def show_image_stack(images, minmax, fontsize=18, cmap='CMRmap',
                     zlabel=r'Intensty [ADU]', figsize=(12, 10)):
    """Show an Interactive Image Stack in an IPython Notebook

    Parameters
    ----------
    images : array_like
        Stack of images of shape (N, y, x) where N is the number of images
        to show.
    minmax : tuple
        Value for the minimum and maximum of the stack in the form
        ``(min, max)``
    fontsize : int
        Fontsize for axis labels.
    cmap : string
        Colormap to use for image (from matplotlib)
    zlabel : string
        Axis label for the color bar (z-axis)
    figsize : tuple
        Figure size (from matplotlib)


    """
    n = images.shape[0]

    def view_frame(i, vmin, vmax):
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

        im = ax.imshow(images[i], cmap=cmap, interpolation='none',
                       vmin=vmin, vmax=vmax)

        cbar = fig.colorbar(im)
        cbar.ax.tick_params(labelsize=fontsize)
        cbar.set_label(zlabel, size=fontsize, weight='bold')

        ax.set_title('Frame {} Min = {} Max = {}'.format(i, vmin, vmax),
                     fontsize=fontsize, fontweight='bold')

        for item in ([ax.xaxis.label, ax.yaxis.label] +
                     ax.get_xticklabels() + ax.get_yticklabels()):
            item.set_fontsize(fontsize)
            item.set_fontweight('bold')

        plt.show()

    interact(view_frame, i=(0, n-1), vmin=minmax, vmax=minmax)


def image_stack_to_movie(images, frames=None, vmin=None, vmax=None,
                         figsize=(6, 5), cmap='CMRmap', fps=10):
    """Convert image stack to movie and show in notebook.

    Parameters
    ----------
    images : array_like
        Stack of images to show as a movie of shape (N, y, x).
    frames : int
        Number of frames to process
    vmin : number
        Minimum value to display for ``imshow``
    vmax : number
        Maximum value to display for ``imshow``
    figsize : tuple
        Figure size for each frame
    cmap : string
        Colormap to use for plotting image
    fps : int
        Framerate for created movie

    """
    if frames is None:
        frames = images.shape[0]

    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    im = plt.imshow(images[1], vmin=vmin, vmax=vmax, cmap=cmap,
                    interpolation='none')
    cbar = fig.colorbar(im)
    cbar.ax.tick_params(labelsize=14)
    cbar.set_label(r"Intensity [ADU]", size=14,)
    for item in ([ax.xaxis.label, ax.yaxis.label] +
                 ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(14)
        item.set_fontweight('bold')

    def animate(i):
        im.set_array(images[i])
        ax.set_title('Frame {}'.format(i), fontsize=16, fontweight='bold')
        return im,

    anim = animation.FuncAnimation(fig, animate, frames=frames,
                                   interval=1, blit=True)
    plt.close(anim._fig)
    # return anim.to_html5_video()
    return HTML(_anim_to_html(anim, fps))


def _anim_to_html(anim, fps):
    VIDEO_TAG = """<video autoplay loop controls>
    <source src="data:video/x-m4v;base64,{0}" type="video/mp4">
    Your browser does not support the video tag.
    </video>"""

    if not hasattr(anim, '_encoded_video'):
        with NamedTemporaryFile(suffix='.mp4') as f:
            anim.save(f.name, fps=fps,
                      extra_args=['-vcodec', 'libx264',
                                  '-pix_fmt', 'yuv420p'])
            video = open(f.name, "rb").read()
        anim._encoded_video = base64.b64encode(video)
    return VIDEO_TAG.format(anim._encoded_video.decode("utf-8"))



from IPython.display import display, HTML

_js_callback_open = """
function callback(out){
    window.open(out.content.text,'_blank');
}
"""

_js = """
var nb = IPython.notebook;
var kernel = IPython.notebook.kernel;
kernel.execute('import subprocess, os')

nb.save_checkpoint();

function gistit(){
    var command = "print("
    command += "subprocess.run(['gist',";
    command += "'"+nb.base_url+nb.notebook_path+"'.replace('/user','/home')]";
    command += ", env=dict(os.environ, http_proxy='http://proxy:8888')";
    command += ", stdout=subprocess.PIPE).stdout.decode('utf-8').rstrip()";
    command += ".replace('gist.github.com','nbviewer.jupyter.org/gist')"
    command += ")";
    kernel.execute(command, {iopub: {output: callback}});
}

setTimeout(gistit, 20000);
"""


def notebook_to_nbviewer():
    js = _js_callback_open + _js
    html = '<script type="text/javascript">'
    html += js
    html += '</script>'
    html += '<b>nbviewer will open in a new tab in 20 seconds .....</b>'
    return display(HTML(html))



__all__ = ['image_stack_to_movie', 'show_image_stack',
           'notebook_to_nbviewer', 'get_sys_info',
           'show_kernels']

from .animation import image_stack_to_movie, show_image_stack
from .nbviewer import notebook_to_nbviewer
from .info import get_sys_info, show_kernels



#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup, ioc_arg_parser, run
import numpy as np
import time
from textwrap import dedent


class Thermo(PVGroup):
    """
    Simulates (poorly) an oscillating temperature controller.

    Follows :math:`T_{output} = T_{var} exp^{-(t - t_0)/K} sin(ω t)
                   + T_{setpoint}`

    The default prefix is `thermo:`

    Readonly PVs
    ------------

    I -> the current value

    Control PVs
    -----------

    SP -> where to go
    K -> the decay constant
    omega -> the oscillation frequency
    Tvar -> the scale of the oscillations
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._T0 = time.monotonic()

    readback = pvproperty(value=0, dtype=float, read_only=True,
                          name='T-RB',
                          mock_record='ai')

    setpoint = pvproperty(value=100, dtype=float, name='T-SP')
    K = pvproperty(value=10, dtype=float)
    omega = pvproperty(value=np.pi, dtype=float)
    Tvar = pvproperty(value=10, dtype=float)

    @readback.scan(period=.1, use_scan_field=True)
    async def readback(self, instance, async_lib):

        def t_rbv(T0, setpoint, K, omega, Tvar,):
            t = time.monotonic()
            return ((Tvar *
                     np.exp(-(t - self._T0) / K) *
                     np.sin(omega * t)) +
                    setpoint)

        T = t_rbv(T0=self._T0,
                  **{k: getattr(self, k).value
                     for k in ['setpoint', 'K', 'omega', 'Tvar']})

        await instance.write(value=T)

    @setpoint.putter
    async def setpoint(self, instance, value):
        self._T0 = time.monotonic()
        return value


if __name__ == '__main__':
    ioc_options, run_options = ioc_arg_parser(
        default_prefix='thermo:',
        desc=dedent(Thermo.__doc__))
    ioc = Thermo(**ioc_options)
    run(ioc.pvdb, **run_options)



#!/usr/bin/env python3
from caproto.server import pvproperty, PVGroup
from caproto.server import template_arg_parser, run
from caproto import ChannelType
import contextvars
import functools

internal_process = contextvars.ContextVar('internal_process',
                                          default=False)


def no_reentry(func):
    @functools.wraps(func)
    async def inner(*args, **kwargs):
        if internal_process.get():
            return
        try:
            internal_process.set(True)
            return (await func(*args, **kwargs))
        finally:
            internal_process.set(False)

    return inner


class EPSTwoStateIOC(PVGroup):
    """
    Simulates an EPS Two State device including multiple-attempt issue
    for two-state device.

    This IOC is used to simulate an EPS Two State Device, for testing
    or development  purposes. Known EPS two state devices include Photon
    shutters, gate valves and Pneumatic actuators at NSLS-II. It simulates
    known issues with some of these including: A hardware error (when some
    attempts at sending the command fail and we need to 'kick' the device
    a few times to get it to actuate), A position status error (it sometimes
    does not reach the final state but remains 'between states') and
    an enable state change error (there is an 'enable' PV controlled by
    the control room that determines if the device can be operated or not).
    A parameter (described below) allows each of these error paths to be
    tested against.

    Parameters
    ----------
    retries : int, optional
        Number of attempts required for changing state,
        default is 2
    enbl_sts_val: str, optional
        Enumerated string that enables the state change,
        default is True
    hw_error_val: str, optional
        Enumerated string that activates the hardware error,
        default is False
    sts_error_val: str, optional
        Enumerated string that activates the Pos-Sts error,
        default is False
    """

    def __init__(self, retries=2, enbl_sts_val='True',
                 hw_error_val='False', sts_error_val='False',
                 **kwargs):

        super().__init__(**kwargs)

        self._max_retries = retries
        self._num_retries = 0

        self._enbl_sts_val = enbl_sts_val

        self._hw_error_val = hw_error_val
        self._sts_error_val = sts_error_val

    # Pos-Sts two-state PV

    _pos_states = ['Open', 'Not Open']  # two position states

    pos_sts = pvproperty(value="Open",
                         enum_strings=_pos_states,
                         dtype=ChannelType.ENUM,
                         read_only=True,
                         name='Pos-Sts')

    # Opn-Cmd and Cls-Cmd PVs used by client for changing state

    _cmd_states = ['None', 'Done']  # two command states

    state1_cmd = pvproperty(value=_cmd_states[0],
                            enum_strings=['None', 'Open'],
                            dtype=ChannelType.ENUM,
                            name='Cmd:Opn-Cmd')
    state2_cmd = pvproperty(value=_cmd_states[0],
                            enum_strings=['None', 'Close'],
                            dtype=ChannelType.ENUM,
                            name='Cmd:Cls-Cmd')

    _fail_states = ['False', 'True']

    fail_to_state1 = pvproperty(value=_fail_states[0],
                                enum_strings=_fail_states,
                                dtype=ChannelType.ENUM,
                                read_only=True,
                                name='Sts:FailOpn-Sts')
    fail_to_state2 = pvproperty(value=_fail_states[0],
                                enum_strings=_fail_states,
                                dtype=ChannelType.ENUM,
                                read_only=True,
                                name='Sts:FailCls-Sts')

    # Enbl-Sts PV that enables/disables the state change

    _enbl_states = ['False', 'True']

    enbl_sts = pvproperty(value='',
                          enum_strings=_enbl_states,
                          dtype=ChannelType.ENUM,
                          name='Enbl-Sts')

    # Hardware error status

    _hw_error_states = ['False', 'True']

    hw_error_sts = pvproperty(value='',
                              enum_strings=_hw_error_states,
                              dtype=ChannelType.ENUM,
                              name='HwError-Sts')

    # Pos-Sts error status

    _sts_error_states = ['False', 'True']

    sts_error_sts = pvproperty(value='',
                               enum_strings=_sts_error_states,
                               dtype=ChannelType.ENUM,
                               name='StsError-Sts')

    # PV Startup/Putter Methods

    @enbl_sts.startup
    async def enbl_sts(self, instance, async_lib):
        await instance.write(value=self._enbl_sts_val)

    @hw_error_sts.startup
    async def hw_error_sts(self, instance, async_lib):
        await instance.write(value=self._hw_error_val)

    @sts_error_sts.startup
    async def sts_error_sts(self, instance, async_lib):
        await instance.write(value=self._sts_error_val)

    @state1_cmd.startup
    async def state1_cmd(self, instance, async_lib):
        instance.async_lib = async_lib

    @state1_cmd.putter
    @no_reentry
    async def state1_cmd(self, instance, value):
        if value == 'Open':
            await self.state1_cmd.write(value)
            await instance.async_lib.library.sleep(1)
            await self.pos_sts.write('Open')
        return 'None'

    @state2_cmd.startup
    async def state2_cmd(self, instance, async_lib):
        instance.async_lib = async_lib

    @state2_cmd.putter
    @no_reentry
    async def state2_cmd(self, instance, value):
        if value == 'Close':
            await self.state2_cmd.write(value)
            await instance.async_lib.library.sleep(1)
            await self.pos_sts.write('Not Open')
        return 'None'

    @enbl_sts.putter
    async def enbl_sts(self, instance, value):
        self._enbl_sts_val = value
        return value

    @hw_error_sts.putter
    async def hw_error_sts(self, instance, value):
        self._hw_error_val = value
        return value

    @sts_error_sts.putter
    async def sts_error_sts(self, instance, value):
        self._sts_error_val = value
        return value

    # Internal Methods

    async def _state_cmd_put(self, instance, value, state_val, fail_to_state):
        if(value == self._cmd_states[0]):  # if None -> do nothing
            return self._cmd_states[0]
        if(self._pos_sts_val == state_val):  # if in state -> do nothing
            return self._cmd_states[0]
        if(self._enbl_sts_val == 'False'):  # if changes not enabled -> fail
            await fail_to_state.write(value='True')
            return self._cmd_states[0]
        self._num_retries += 1
        if(self._num_retries < self._max_retries):
            return self._cmd_states[0]
        else:
            self._num_retries = 0
        if(self._hw_error_val == 'True'):  # if hw error -> fail
            await fail_to_state.write(value='True')
            return self._cmd_states[1]
        else:
            await fail_to_state.write(value='False')
        if(self._sts_error_val == 'True'):  # if sts error -> don't change sts
            return self._cmd_states[1]
        await self.pos_sts.write(value=state_val)
        self._pos_sts_val = state_val
        return self._cmd_states[0]


if __name__ == '__main__':

    parser, split_args = template_arg_parser(
        default_prefix='eps2state:',
        desc='EPS Two State IOC.')

    retries_help = 'Number of attempts required for changing state'
    enable_help = 'State change is enabled'
    hwerror_help = 'HW error is activated'
    stserror_help = 'Pos-Sts error is activated'

    parser.add_argument('--retries', help=retries_help,
                        required=False, default=2, type=int)
    parser.add_argument('--enable', help=enable_help,
                        required=False, default='True', type=str)
    parser.add_argument('--hwerror', help=hwerror_help,
                        required=False, default='False', type=str)
    parser.add_argument('--stserror', help=stserror_help,
                        required=False, default='False', type=str)

    args = parser.parse_args()
    ioc_options, run_options = split_args(args)

    ioc = EPSTwoStateIOC(retries=args.retries,
                         enbl_sts_val=args.enable,
                         hw_error_val=args.hwerror,
                         sts_error_val=args.stserror,
                         **ioc_options)

    run(ioc.pvdb, **run_options)



from distutils.version import LooseVersion
import logging
from logging.handlers import SysLogHandler, TimedRotatingFileHandler
import os
from pathlib import Path
import sys
import warnings

import appdirs

from IPython import get_ipython

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


bluesky_log_file_path = None


def import_star(module, ns):
    def public(name):
        return not name.startswith("_")

    ns.update({name: getattr(module, name) for name in dir(module) if public(name)})


def configure_base(
    user_ns,
    broker_name,
    *,
    bec=True,
    bec_derivative=False,
    epics_context=False,
    magics=True,
    mpl=True,
    configure_logging=True,
    pbar=True,
    ipython_logging=True,
    publish_documents_with_kafka=False,
    tb_minimize=True,
):
    """
    Perform base setup and instantiation of important objects.

    This factory function instantiates essential objects to data collection
    environments at NSLS-II and adds them to the current namespace. In some
    cases (documented below), it will check whether certain variables already
    exist in the user name space, and will avoid creating them if so. The
    following are added:

    * ``RE`` -- a RunEngine
        This is created only if an ``RE`` instance does not currently exist in
        the namespace.
    * ``db`` -- a Broker (from "databroker"), subscribe to ``RE``
    * ``bec`` -- a BestEffortCallback, subscribed to ``RE``
    * ``peaks`` -- an alias for ``bec.peaks``
    * ``sd`` -- a SupplementalData preprocessor, added to ``RE.preprocessors``
    * ``pbar_maanger`` -- a ProgressBarManager, set as the ``RE.waiting_hook``

    And it performs some low-level configuration:

    * creates a context in ophyd's control layer (``ophyd.setup_ophyd()``)
    * turns on interactive plotting (``matplotlib.pyplot.ion()``)
    * bridges the RunEngine and Qt event loops
      (``bluesky.utils.install_kicker()``)
    * logs ERROR-level log message from ophyd to the standard out

    Parameters
    ----------
    user_ns: dict
        a namespace --- for example, ``get_ipython().user_ns``
    broker_name : Union[str, Broker]
        Name of databroker configuration or a Broker instance.
    bec : boolean, optional
        True by default. Set False to skip BestEffortCallback.
    bec_derivative : boolean, optional
        False by default. Set True to enable derivative and its stats
        calculation in BestEffortCallback.
    epics_context : boolean, optional
        True by default. Set False to skip ``setup_ophyd()``.
    magics : boolean, optional
        True by default. Set False to skip registration of custom IPython
        magics.
    mpl : boolean, optional
        True by default. Set False to skip matplotlib ``ion()`` at event-loop
        bridging.
    configure_logging : boolean, optional
        True by default. Set False to skip INFO-level logging.
    pbar : boolean, optional
        True by default. Set false to skip ProgressBarManager.
    ipython_logging : boolean, optional
        True by default. Console output and exception stack traces will be
        written to IPython log file when IPython logging is enabled.
    publish_documents_with_kafka: boolean, optional
        False by default. If True publish bluesky documents to a Kafka message broker using
        configuration parameters read from a file.
    tb_minimize : boolean, optional
        If IPython should print out 'minimal' tracebacks.

    Returns
    -------
    names : list
        list of names added to the namespace

    Examples
    --------
    Configure IPython for CHX.

    >>>> configure_base(get_ipython().user_ns, 'chx');
    """

    ipython = get_ipython()

    ns = {}  # We will update user_ns with this at the end.
    # Protect against double-subscription.
    SENTINEL = "__nslsii_configure_base_has_been_run"
    if user_ns.get(SENTINEL):
        raise RuntimeError("configure_base should only be called once per process.")
    ns[SENTINEL] = True
    # Set up a RunEngine and use metadata backed by files on disk.
    from bluesky import RunEngine, __version__ as bluesky_version

    if LooseVersion(bluesky_version) >= LooseVersion("1.6.0"):
        # current approach using PersistentDict
        from bluesky.utils import PersistentDict

        directory = os.path.expanduser("~/.config/bluesky/md")
        os.makedirs(directory, exist_ok=True)
        md = PersistentDict(directory)
    else:
        # legacy approach using HistoryDict
        from bluesky.utils import get_history

        md = get_history()
    # if RunEngine already defined grab it
    # useful when users make their own custom RunEngine
    if "RE" in user_ns:
        RE = user_ns["RE"]
    else:
        RE = RunEngine(md)
        ns["RE"] = RE

    # Set up SupplementalData.
    # (This is a no-op until devices are added to it,
    # so there is no need to provide a 'skip_sd' switch.)
    from bluesky import SupplementalData

    sd = SupplementalData()
    RE.preprocessors.append(sd)
    ns["sd"] = sd

    if isinstance(broker_name, str):
        # Set up a Broker.
        from databroker import Broker

        db = Broker.named(broker_name)
        ns["db"] = db
    else:
        db = broker_name

    RE.subscribe(db.insert)

    if pbar:
        # Add a progress bar.
        from bluesky.utils import ProgressBarManager

        pbar_manager = ProgressBarManager()
        RE.waiting_hook = pbar_manager
        ns["pbar_manager"] = pbar_manager

    if magics:
        # Register bluesky IPython magics.
        from bluesky.magics import BlueskyMagics

        if ipython:
            ipython.register_magics(BlueskyMagics)

    if bec:
        # Set up the BestEffortCallback.
        from bluesky.callbacks.best_effort import BestEffortCallback

        _bec_kwargs = {}
        if bec_derivative:
            _bec_kwargs["calc_derivative_and_stats"] = True

        _bec = BestEffortCallback(**_bec_kwargs)
        RE.subscribe(_bec)
        ns["bec"] = _bec
        ns["peaks"] = _bec.peaks  # just as alias for less typing

    if mpl:
        # Import matplotlib and put it in interactive mode.
        import matplotlib.pyplot as plt

        ns["plt"] = plt
        plt.ion()

        # Make plots update live while scans run.
        if LooseVersion(bluesky_version) < LooseVersion("1.6.0"):
            from bluesky.utils import install_kicker

            install_kicker()

    if epics_context:
        # Create a context in the underlying EPICS client.
        from ophyd import setup_ophyd

        setup_ophyd()

    if configure_logging:
        configure_bluesky_logging(ipython=ipython)

    if ipython_logging and ipython:
        from nslsii.common.ipynb.logutils import log_exception

        # IPython logging will be enabled with logstart(...)
        configure_ipython_logging(exception_logger=log_exception, ipython=ipython)

    if publish_documents_with_kafka:
        configure_kafka_publisher(RE, beamline_name=broker_name)

    if tb_minimize and ipython:
        # configure %xmode minimal
        # so short tracebacks are printed to the console
        ipython.run_line_magic("xmode", "minimal")

    # convenience imports
    # some of the * imports are for 'back-compatibility' of a sort -- we have
    # taught BL staff to expect LiveTable and LivePlot etc. to be in their
    # namespace
    import numpy as np

    ns["np"] = np

    import bluesky.callbacks

    ns["bc"] = bluesky.callbacks
    import_star(bluesky.callbacks, ns)

    import bluesky.plans

    ns["bp"] = bluesky.plans
    import_star(bluesky.plans, ns)

    import bluesky.plan_stubs

    ns["bps"] = bluesky.plan_stubs
    import_star(bluesky.plan_stubs, ns)
    # special-case the commonly-used mv / mvr and its aliases mov / movr4
    ns["mv"] = bluesky.plan_stubs.mv
    ns["mvr"] = bluesky.plan_stubs.mvr
    ns["mov"] = bluesky.plan_stubs.mov
    ns["movr"] = bluesky.plan_stubs.movr

    import bluesky.preprocessors

    ns["bpp"] = bluesky.preprocessors

    import bluesky.callbacks.broker

    import_star(bluesky.callbacks.broker, ns)

    import bluesky.simulators

    import_star(bluesky.simulators, ns)

    user_ns.update(ns)
    return list(ns)


def configure_bluesky_logging(
    ipython, appdirs_appname="bluesky", propagate_log_messages=False
):
    """
    Configure a TimedRotatingFileHandler log handler and attach it to
    bluesky, ophyd, caproto, and nslsii loggers. In addition, by default set
    the ``propagate`` field on each logger to ``False`` so log messages will
    not propagate to higher level loggers such as a root logger configured
    by a user. If you want log messages from these loggers to propagate to
    higher level loggers simply set ``propagate_log_messages=True`` when
    calling this function, or set the ``propagate`` field to ``True`` in
    client code.

    The log file path is taken from environment variable BLUESKY_LOG_FILE, if
    that variable has been set. If not the default log file location is determined
    by the appdirs package. The default log directory will be created if it does
    not exist.

    Parameters
    ----------
    ipython: InteractiveShell
        IPython InteractiveShell used to attach bluesky log handler to ipython
    appdirs_appname: str
        appname passed to appdirs.user_log_dir() when the BLUESKY_LOG_FILE
        environment variable has not been set; use the default for production,
        set to something else for testing
    propagate_log_messages: bool
        the ``propagate`` field on the bluesky, caproto, nslsii, ophyd, and ipython
        loggers will be set to this value; if False (the default) log messages
        from these loggers will not propagate to higher-level loggers
        (such as a root logger)

    Returns
    -------
    bluesky_log_file_path: Path
        log file path

    """
    global bluesky_log_file_path

    if "BLUESKY_LOG_FILE" in os.environ:
        bluesky_log_file_path = Path(os.environ["BLUESKY_LOG_FILE"])
        print(
            f"bluesky log file path configured from environment variable"
            f" BLUESKY_LOG_FILE: '{bluesky_log_file_path}'",
            file=sys.stderr,
        )
    else:
        bluesky_log_dir = Path(appdirs.user_log_dir(appname=appdirs_appname))
        if not bluesky_log_dir.exists():
            bluesky_log_dir.mkdir(parents=True, exist_ok=True)
        bluesky_log_file_path = bluesky_log_dir / Path("bluesky.log")
        print(
            f"environment variable BLUESKY_LOG_FILE is not set,"
            f" using default log file path '{bluesky_log_file_path}'",
            file=sys.stderr,
        )

    logging_handlers = []

    log_file_handler = TimedRotatingFileHandler(
        filename=str(bluesky_log_file_path), when="W0", backupCount=10
    )
    log_file_handler.setLevel("INFO")
    log_file_format = (
        "[%(levelname)1.1s %(asctime)s.%(msecs)03d %(name)s"
        "  %(module)s:%(lineno)d] %(message)s"
    )
    log_file_handler.setFormatter(logging.Formatter(fmt=log_file_format))
    logging_handlers.append(log_file_handler)

    def build_syslog_handler(address):
        syslog_handler = SysLogHandler(address=address)
        syslog_handler.setLevel(logging.INFO)
        # no need to log date and time, systemd does that
        formatter = logging.Formatter(
            "%(name)s[%(process)s]: %(levelname)s - %(module)s:%(lineno)d] %(message)s"
        )
        # add formatter to syslog_handler
        syslog_handler.setFormatter(formatter)
        return syslog_handler

    if Path("/dev/log").exists():
        logging_handlers.append(build_syslog_handler(address="/dev/log"))
    elif Path("/var/run/syslog").exists():
        logging_handlers.append(build_syslog_handler(address="/var/run/syslog"))
    else:
        # syslog is not available available
        pass

    for logger_name in ("bluesky", "caproto", "ophyd", "nslsii"):
        logger = logging.getLogger(logger_name)
        for logging_handler in logging_handlers:
            logger.addHandler(logging_handler)
        logger.setLevel("INFO")
        logger.propagate = propagate_log_messages

    if ipython:
        for logging_handler in logging_handlers:
            ipython.log.addHandler(logging_handler)

        ipython.log.setLevel("INFO")
        ipython.log.propagate = propagate_log_messages

    return bluesky_log_file_path


def configure_ipython_logging(
    exception_logger, ipython, rotate_file_size=100000, appdirs_appname="bluesky"
):
    """
    Configure IPython output logging with logstart and IPython exception logging with set_custom_exc(...).

    Set a custom exception logging function and execute logstart.

    The log file path is taken from environment variable BLUESKY_IPYTHON_LOG_FILE, if
    it that variable has been set. If not the default log file location is determined
    by the appdirs package.

    Parameters
    ----------
    exception_logger: function f(ipyshell, etype, evalue, tb, tb_offset=None) -> list
        a function that will handle logging exceptions
    ipython: InteractiveShell
        IPython InteractiveShell into which the specified exception_logger will be installed
    rotate_file_size: int, optional
        at the time configure_ipython_exc_logging() is called, if there exists a log file
        with size in bytes greater than or equal to rotate_file_size, the existing file will
        be renamed and a new log file will be created
    appdirs_appname: str
        appname passed to appdirs.user_log_dir(); use the default for production,
        set to something else for testing

    Returns
    -------
    bluesky_ipython_log_file_path: Path
        log file path

    """
    # install the specified function to log exceptions
    ipython.set_custom_exc((BaseException,), exception_logger)

    if "BLUESKY_IPYTHON_LOG_FILE" in os.environ:
        bluesky_ipython_log_file_path = Path(os.environ["BLUESKY_IPYTHON_LOG_FILE"])
        print(
            "bluesky ipython log file configured from environment"
            f" variable BLUESKY_IPYTHON_LOG_FILE: '{bluesky_ipython_log_file_path}'",
            file=sys.stderr,
        )
    else:
        bluesky_ipython_log_dir = Path(appdirs.user_log_dir(appname=appdirs_appname))
        if not bluesky_ipython_log_dir.exists():
            bluesky_ipython_log_dir.mkdir(parents=True, exist_ok=True)
        bluesky_ipython_log_file_path = bluesky_ipython_log_dir / Path(
            "bluesky_ipython.log"
        )
        print(
            "environment variable BLUESKY_IPYTHON_LOG_FILE is not set,"
            f" using default file path '{bluesky_ipython_log_file_path}'",
            file=sys.stderr,
        )
    # before starting ipython logging check the size of the ipython log file
    # if the ipython log file has grown large make a copy and start a new one
    # if a previous copy exists just overwrite it
    if (
        bluesky_ipython_log_file_path.exists()
        and os.path.getsize(bluesky_ipython_log_file_path) >= rotate_file_size
    ):
        bluesky_ipython_log_file_path.rename(
            str(bluesky_ipython_log_file_path) + ".old"
        )
    # ipython gives a warning if logging fails to start, for example if the log
    # directory does not exist. Convert that warning to an exception here.
    with warnings.catch_warnings():
        warnings.simplefilter(action="error")
        # specify the file for ipython logging output
        ipython.run_line_magic("logstart", f"-o -t {bluesky_ipython_log_file_path} append")

    return bluesky_ipython_log_file_path


def configure_kafka_publisher(RE, beamline_name, override_config_path=None):
    """ Read a Kafka configuration file and subscribe a Kafka publisher to the RunEngine.

    A configuration file is required. Environment variable BLUESKY_KAFKA_CONFIG_FILE
    will be checked if `configuration_file_path` is not specified. Otherwise the default
    path `/etc/bluesky/kafka.yml` will be read.

    The intention is that the default path is used in production. The environment variable
    allows for modifying a deployed system and the parameter is useful for testing.

    See `tests/test_kafka_configuration.py` for an example configuration file.
    """
    from nslsii.kafka_utils import (
        _read_bluesky_kafka_config_file,
        _subscribe_kafka_publisher,
        _subscribe_kafka_queue_thread_publisher
    )

    bluesky_kafka_config_path = None
    kafka_publisher_details = None

    if override_config_path is not None:
        bluesky_kafka_config_path = override_config_path
    elif "BLUESKY_KAFKA_CONFIG_PATH" in os.environ:
        bluesky_kafka_config_path = os.environ["BLUESKY_KAFKA_CONFIG_PATH"]
    else:
        bluesky_kafka_config_path = "/etc/bluesky/kafka.yml"

    bluesky_kafka_configuration = _read_bluesky_kafka_config_file(
        bluesky_kafka_config_path
    )
    # convert the list of bootstrap servers into a comma-delimited string
    #   which is the format required by the confluent python api
    bootstrap_servers = ",".join(bluesky_kafka_configuration["bootstrap_servers"])

    runengine_producer_config = {}
    if "producer_consumer_security_config" in bluesky_kafka_configuration:
        runengine_producer_config.update(bluesky_kafka_configuration["producer_consumer_security_config"])
    runengine_producer_config.update(bluesky_kafka_configuration["runengine_producer_config"])

    if bluesky_kafka_configuration["abort_run_on_kafka_exception"]:
        kafka_publisher_details = _subscribe_kafka_publisher(
            RE,
            beamline_name=beamline_name,
            bootstrap_servers=bootstrap_servers,
            producer_config=bluesky_kafka_configuration["runengine_producer_config"]
        )
    else:
        kafka_publisher_details = _subscribe_kafka_queue_thread_publisher(
            RE,
            beamline_name=beamline_name,
            bootstrap_servers=bootstrap_servers,
            producer_config=bluesky_kafka_configuration[
                "runengine_producer_config"
            ],
        )

    return bluesky_kafka_configuration, kafka_publisher_details


def configure_olog(user_ns, *, callback=None, subscribe=True):
    """
    Setup a callback that publishes some metadata from the RunEngine to Olog.

    Also, add the public contents of pyOlog.ophyd_tools to the namespace.

    This is expected to be run after :func:`configure_base`. It expects to find
    an instance of RunEngine named ``RE`` in the user namespace. Additionally,
    if the user namespace contains the name ``logbook``, that is expected to be
    an instance ``pyOlog.SimpleOlogClient``.

    Parameters
    ----------
    user_ns: dict
        a namespace --- for example, ``get_ipython().user_ns``
    callback : callable, optional
        a hook for customizing the logbook_cb_factory; if None a default is
        used
    subscribe : boolean, optional
        True by default. Set to False to skip the subscription. (You still get
        pyOlog.ophyd_tools.)

    Returns
    -------
    names : list
        list of names added to the namespace

    Examples
    --------
    Configure the Olog.

    >>>> configure_olog(get_ipython().user_ns);
    """
    # Conceptually our task is simple: add a subscription to the RunEngine that
    # publishes to the Olog using the Python wrapper of its REST API, pyOlog.
    # In practice this is messy because we have deal with the many-layered API
    # of pyOlog and, more importantly, ensure that slowness or errors from the
    # Olog do not affect the run. Historically the Olog deployment has not been
    # reliable, so it is important to be robust against these issues. Of
    # course, by ignoring Olog errors, we leave gaps in the log, which is not
    # great, but since all data is saved to a databroker anyway, we can always
    # re-generate them later.

    ns = {}  # We will update user_ns with this at the end.

    from bluesky.callbacks.olog import logbook_cb_factory
    from functools import partial
    from pyOlog import SimpleOlogClient
    import queue
    import threading
    from warnings import warn

    # This is for pyOlog.ophyd_tools.get_logbook, which simply looks for
    # a variable called 'logbook' in the global IPython namespace.
    if "logbook" in user_ns:
        simple_olog_client = user_ns["logbook"]
    else:
        simple_olog_client = SimpleOlogClient()
        ns["logbook"] = simple_olog_client

    if subscribe:
        if callback is None:
            # list of logbook names to publish to
            LOGBOOKS = ("Data Acquisition",)
            generic_logbook_func = simple_olog_client.log
            configured_logbook_func = partial(generic_logbook_func, logbooks=LOGBOOKS)
            callback = logbook_cb_factory(configured_logbook_func)

        def submit_to_olog(queue, cb):
            while True:
                name, doc = queue.get()  # waits until document is available
                try:
                    cb(name, doc)
                except Exception as exc:
                    warn(
                        "This olog is giving errors. This will not be logged."
                        "Error:" + str(exc)
                    )

        olog_queue = queue.Queue(maxsize=100)
        olog_thread = threading.Thread(
            target=submit_to_olog, args=(olog_queue, callback), daemon=True
        )

        olog_thread.start()

        def send_to_olog_queue(name, doc):
            try:
                olog_queue.put((name, doc), block=False)
            except queue.Full:
                warn("The olog queue is full. This will not be logged.")

        RE = user_ns["RE"]
        RE.subscribe(send_to_olog_queue, "start")

    import pyOlog.ophyd_tools

    import_star(pyOlog.ophyd_tools, ns)

    user_ns.update(ns)
    return list(ns)


def migrate_metadata():
    """
    Copy metadata from (old) sqlite-backed file to (new) directory of msgpack.
    """
    from bluesky.utils import get_history, PersistentDict

    old_md = get_history()
    directory = os.path.expanduser("~/.config/bluesky/md")
    os.makedirs(directory, exist_ok=True)
    new_md = PersistentDict(directory)
    new_md.update(old_md)



# DataBroker "transforms" for patching up documents
# on their way out.
import copy
import os


def csx_fix_scaler_shape(d):
    """
    Transform an Event Descriptor.

    Look for sources that match the pattern of a Scaler.
    Ensure that their shape is [].
    """
    d = copy.deepcopy(d)
    for k, v in list(d['data_keys'].items()):
        if v['source'].startswith('PV:XF:23ID1-ES{Sclr:1}Wfrm'):
            d['data_keys'][k]['shape'] = []
    return d

def srx_transform_resource(doc):
    """
    This patch is used to update the root and resource path of resource documents
    so that the Filler looks for the files on the Lustre instead of GPFS.  This is
    needed because root_map is not sufficient in this case.
    """
    doc = dict(doc)
    full_path = os.path.join(doc['root'], doc['resource_path'])
    new_path = full_path.replace('/nsls2/xf05id1/XF05ID1', '/nsls2/data/srx/legacy/xf05id1/XF05ID1')
    doc['root'] = ''
    doc['resource_path'] = new_path

    return doc



import logging
import sys
import traceback

from nslsii import bluesky_log_file_path


def log_exception(ipyshell, etype, evalue, tb, tb_offset=None):
    """A custom IPython exception handler that logs exception tracebacks to
    the IPython log file as well as to the nslsii.ipython logger.



    References:
        https://ipython.readthedocs.io/en/stable/api/generated/IPython.core.interactiveshell.html
        https://ipython.readthedocs.io/en/stable/api/generated/IPython.core.logger.html

    Usage:
        from nslsii.common.ipynb.logutils import log_exception
        get_ipython().set_custom_exc((BaseException, ), log_exception)

        %logstart -o -t ipython_log/beamline.log append

    Parameters
    ----------
    ipyshell : TerminalInteractiveShell

    etype : type of evalue

    evalue : BaseException

    tb : traceback

    tb_offset : ???

    Returns
    -------
    list of traceback lines
    """

    # send the traceback to the IPython log file
    tb_lines = traceback.format_exception(etype, evalue, tb)
    for tb_line in tb_lines:
        ipyshell.logger.log_write(tb_line, kind="output")

    # display the exception in the console
    if ipyshell.InteractiveTB.mode == "Minimal":
        print(
            "An exception has occurred, use '%tb verbose'"
            " to see the full traceback.",
            file=sys.stderr,
        )
    ipyshell.showtraceback((etype, evalue, tb), tb_offset=tb_offset)

    # send the traceback to the nslsii.ipython logger
    logging.getLogger("nslsii.ipython").exception(evalue)
    print(
        f"See {bluesky_log_file_path} for the full traceback.",
        file=sys.stderr
    )

    return tb_lines



import time
import datetime
from ophyd import (Device, Component as Cpt,
                   EpicsSignal, EpicsSignalRO, DeviceStatus)

_time_fmtstr = '%Y-%m-%d %H:%M:%S'


class TwoButtonShutter(Device):
    # TODO: this needs to be fixed in EPICS as these names make no sense
    # the value coming out of the PV does not match what is shown in CSS
    RETRY_PERIOD = 0.5
    MAX_ATTEMPTS = 10
    open_cmd = Cpt(EpicsSignal, 'Cmd:Opn-Cmd', string=True)
    open_val = 'Open'

    close_cmd = Cpt(EpicsSignal, 'Cmd:Cls-Cmd', string=True)
    close_val = 'Not Open'

    status = Cpt(EpicsSignalRO, 'Pos-Sts', string=True)
    fail_to_close = Cpt(EpicsSignalRO, 'Sts:FailCls-Sts', string=True)
    fail_to_open = Cpt(EpicsSignalRO, 'Sts:FailOpn-Sts', string=True)
    enabled_status = Cpt(EpicsSignalRO, 'Enbl-Sts', string=True)

    # user facing commands
    open_str = 'Open'
    close_str = 'Close'

    def set(self, val):
        if self._set_st is not None:
            raise RuntimeError(f'trying to set {self.name}'
                               ' while a set is in progress')

        cmd_map = {self.open_str: self.open_cmd,
                   self.close_str: self.close_cmd}
        target_map = {self.open_str: self.open_val,
                      self.close_str: self.close_val}

        cmd_sig = cmd_map[val]
        target_val = target_map[val]

        st = DeviceStatus(self)
        if self.status.get() == target_val:
            st._finished()
            return st

        self._set_st = st
        print(self.name, val, id(st))
        enums = self.status.enum_strs

        def shutter_cb(value, timestamp, **kwargs):
            try:
                value = enums[int(value)]
            except (ValueError, TypeError):
                # we are here because value is a str not int
                # just move on
                ...
            if value == target_val:
                self._set_st = None
                self.status.clear_sub(shutter_cb)
                st._finished()

        cmd_enums = cmd_sig.enum_strs
        count = 0

        def cmd_retry_cb(value, timestamp, **kwargs):
            nonlocal count
            try:
                value = cmd_enums[int(value)]
            except (ValueError, TypeError):
                # we are here because value is a str not int
                # just move on
                ...
            count += 1
            if count > self.MAX_ATTEMPTS:
                cmd_sig.clear_sub(cmd_retry_cb)
                self._set_st = None
                self.status.clear_sub(shutter_cb)
                st._finished(success=False)
            if value == 'None':
                if not st.done:
                    time.sleep(self.RETRY_PERIOD)
                    cmd_sig.set(1)

                    ts = datetime.datetime.fromtimestamp(timestamp) \
                        .strftime(_time_fmtstr)
                    if count > 2:
                        msg = '** ({}) Had to reactuate shutter while {}ing'
                        print(msg.format(ts, val if val != 'Close'
                                         else val[:-1]))
                else:
                    cmd_sig.clear_sub(cmd_retry_cb)

        cmd_sig.subscribe(cmd_retry_cb, run=False)
        self.status.subscribe(shutter_cb)
        cmd_sig.set(1)

        return st

    def stop(self, *, success=False):
        import time
        prev_st = self._set_st
        if prev_st is not None:
            while not prev_st.done:
                time.sleep(.1)
        self._was_open = (self.open_val == self.status.get())
        st = self.set('Close')
        while not st.done:
            time.sleep(.5)

    def resume(self):
        import time
        prev_st = self._set_st
        if prev_st is not None:
            while not prev_st.done:
                time.sleep(.1)
        if self._was_open:
            st = self.set('Open')
            while not st.done:
                time.sleep(.5)

    def unstage(self):
        self._was_open = False
        return super().unstage()

    def __init__(self, *args, **kwargs):
        self._was_open = False
        super().__init__(*args, **kwargs)
        self._set_st = None
        self.read_attrs = ['status']



"""Classes to help with supporting AreaDetector 33 (and the
wait_for_plugins functionality)


This is actually adding a mix of functionality from AD 2-2 to 3-3 and
all of these names may change in the future.

"""

from ophyd import Device, Component as Cpt
from ophyd.areadetector.base import (ADBase, ADComponent as ADCpt, ad_group,
                                     EpicsSignalWithRBV as SignalWithRBV)
from ophyd.areadetector.plugins import PluginBase
from ophyd.areadetector.trigger_mixins import TriggerBase, ADTriggerStatus
from ophyd.device import DynamicDeviceComponent as DDC, Staged
from ophyd.signal import (Signal, EpicsSignalRO, EpicsSignal)
from ophyd.quadem import QuadEM

import time as ttime


class V22Mixin(Device):
    ...


class V26Mixin(V22Mixin):
    adcore_version = Cpt(EpicsSignalRO, 'ADCoreVersion_RBV',
                         string=True, kind='config')
    driver_version = Cpt(EpicsSignalRO, 'DriverVersion_RBV',
                         string=True, kind='config')


class V33Mixin(V26Mixin):
    ...


class CamV33Mixin(V33Mixin):
    wait_for_plugins = Cpt(EpicsSignal, 'WaitForPlugins',
                           string=True, kind='config')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs['wait_for_plugins'] = 'Yes'

    def ensure_nonblocking(self):
        self.stage_sigs['wait_for_plugins'] = 'Yes'
        for c in self.parent.component_names:
            cpt = getattr(self.parent, c)
            if cpt is self:
                continue
            if hasattr(cpt, 'ensure_nonblocking'):
                cpt.ensure_nonblocking()


class FilePluginV22Mixin(V22Mixin):
    create_directories = Cpt(EpicsSignal,
                             'CreateDirectory', kind='config')


class SingleTriggerV33(TriggerBase):
    _status_type = ADTriggerStatus

    def __init__(self, *args, image_name=None, **kwargs):
        super().__init__(*args, **kwargs)
        if image_name is None:
            image_name = '_'.join([self.name, 'image'])
        self._image_name = image_name

    def trigger(self):
        "Trigger one acquisition."
        if self._staged != Staged.yes:
            raise RuntimeError("This detector is not ready to trigger."
                               "Call the stage() method before triggering.")

        self._status = self._status_type(self)

        def _acq_done(*args, **kwargs):
            # TODO sort out if anything useful in here
            self._status._finished()

        self._acquisition_signal.put(1, use_complete=True, callback=_acq_done)
        self.dispatch(self._image_name, ttime.time())
        return self._status


class StatsPluginV33(PluginBase):
    """This supports changes to time series PV names in AD 3-3

    Due to https://github.com/areaDetector/ADCore/pull/333
    """
    _default_suffix = 'Stats1:'
    _suffix_re = r'Stats\d:'
    _html_docs = ['NDPluginStats.html']
    _plugin_type = 'NDPluginStats'

    _default_configuration_attrs = (PluginBase._default_configuration_attrs + (
        'centroid_threshold', 'compute_centroid', 'compute_histogram',
        'compute_profiles', 'compute_statistics', 'bgd_width',
        'hist_size', 'hist_min', 'hist_max', 'ts_num_points', 'profile_size',
        'profile_cursor')
    )

    bgd_width = ADCpt(SignalWithRBV, 'BgdWidth')
    centroid_threshold = ADCpt(SignalWithRBV, 'CentroidThreshold')

    centroid = DDC(ad_group(EpicsSignalRO,
                            (('x', 'CentroidX_RBV'),
                             ('y', 'CentroidY_RBV'))),
                   doc='The centroid XY',
                   default_read_attrs=('x', 'y'))

    compute_centroid = ADCpt(SignalWithRBV, 'ComputeCentroid', string=True)
    compute_histogram = ADCpt(SignalWithRBV, 'ComputeHistogram', string=True)
    compute_profiles = ADCpt(SignalWithRBV, 'ComputeProfiles', string=True)
    compute_statistics = ADCpt(SignalWithRBV, 'ComputeStatistics', string=True)

    cursor = DDC(ad_group(SignalWithRBV,
                          (('x', 'CursorX'),
                           ('y', 'CursorY'))),
                 doc='The cursor XY',
                 default_read_attrs=('x', 'y'))

    hist_entropy = ADCpt(EpicsSignalRO, 'HistEntropy_RBV')
    hist_max = ADCpt(SignalWithRBV, 'HistMax')
    hist_min = ADCpt(SignalWithRBV, 'HistMin')
    hist_size = ADCpt(SignalWithRBV, 'HistSize')
    histogram = ADCpt(EpicsSignalRO, 'Histogram_RBV')

    max_size = DDC(ad_group(EpicsSignal,
                            (('x', 'MaxSizeX'),
                             ('y', 'MaxSizeY'))),
                   doc='The maximum size in XY',
                   default_read_attrs=('x', 'y'))

    max_value = ADCpt(EpicsSignalRO, 'MaxValue_RBV')
    max_xy = DDC(ad_group(EpicsSignalRO,
                          (('x', 'MaxX_RBV'),
                           ('y', 'MaxY_RBV'))),
                 doc='Maximum in XY',
                 default_read_attrs=('x', 'y'))

    mean_value = ADCpt(EpicsSignalRO, 'MeanValue_RBV')
    min_value = ADCpt(EpicsSignalRO, 'MinValue_RBV')

    min_xy = DDC(ad_group(EpicsSignalRO,
                          (('x', 'MinX_RBV'),
                           ('y', 'MinY_RBV'))),
                 doc='Minimum in XY',
                 default_read_attrs=('x', 'y'))

    net = ADCpt(EpicsSignalRO, 'Net_RBV')
    profile_average = DDC(ad_group(EpicsSignalRO,
                                   (('x', 'ProfileAverageX_RBV'),
                                    ('y', 'ProfileAverageY_RBV'))),
                          doc='Profile average in XY',
                          default_read_attrs=('x', 'y'))

    profile_centroid = DDC(ad_group(EpicsSignalRO,
                                    (('x', 'ProfileCentroidX_RBV'),
                                     ('y', 'ProfileCentroidY_RBV'))),
                           doc='Profile centroid in XY',
                           default_read_attrs=('x', 'y'))

    profile_cursor = DDC(ad_group(EpicsSignalRO,
                                  (('x', 'ProfileCursorX_RBV'),
                                   ('y', 'ProfileCursorY_RBV'))),
                         doc='Profile cursor in XY',
                         default_read_attrs=('x', 'y'))

    profile_size = DDC(ad_group(EpicsSignalRO,
                                (('x', 'ProfileSizeX_RBV'),
                                 ('y', 'ProfileSizeY_RBV'))),
                       doc='Profile size in XY',
                       default_read_attrs=('x', 'y'))

    profile_threshold = DDC(ad_group(EpicsSignalRO,
                                     (('x', 'ProfileThresholdX_RBV'),
                                      ('y', 'ProfileThresholdY_RBV'))),
                            doc='Profile threshold in XY',
                            default_read_attrs=('x', 'y'))

    set_xhopr = ADCpt(EpicsSignal, 'SetXHOPR')
    set_yhopr = ADCpt(EpicsSignal, 'SetYHOPR')
    sigma_xy = ADCpt(EpicsSignalRO, 'SigmaXY_RBV')
    sigma_x = ADCpt(EpicsSignalRO, 'SigmaX_RBV')
    sigma_y = ADCpt(EpicsSignalRO, 'SigmaY_RBV')
    sigma = ADCpt(EpicsSignalRO, 'Sigma_RBV')
    ts_acquiring = ADCpt(EpicsSignal, 'TS:TSAcquiring')

    ts_centroid = DDC(ad_group(EpicsSignal,
                               (('x', 'TS:TSCentroidX'),
                                ('y', 'TS:TSCentroidY'))),
                      doc='Time series centroid in XY',
                      default_read_attrs=('x', 'y'))

    # ts_control = ADCpt(EpicsSignal, 'TS:TSControl', string=True)
    ts_current_point = ADCpt(EpicsSignal, 'TS:TSCurrentPoint')
    ts_max_value = ADCpt(EpicsSignal, 'TS:TSMaxValue')

    ts_max = DDC(ad_group(EpicsSignal,
                          (('x', 'TS:TSMaxX'),
                           ('y', 'TS:TSMaxY'))),
                 doc='Time series maximum in XY',
                 default_read_attrs=('x', 'y'))

    ts_mean_value = ADCpt(EpicsSignal, 'TS:TSMeanValue')
    ts_min_value = ADCpt(EpicsSignal, 'TS:TSMinValue')

    ts_min = DDC(ad_group(EpicsSignal,
                          (('x', 'TS:TSMinX'),
                           ('y', 'TS:TSMinY'))),
                 doc='Time series minimum in XY',
                 default_read_attrs=('x', 'y'))

    ts_net = ADCpt(EpicsSignal, 'TS:TSNet')
    ts_num_points = ADCpt(EpicsSignal, 'TS:TSNumPoints')
    ts_read = ADCpt(EpicsSignal, 'TS:TSRead')
    ts_sigma = ADCpt(EpicsSignal, 'TS:TSSigma')
    ts_sigma_x = ADCpt(EpicsSignal, 'TS:TSSigmaX')
    ts_sigma_xy = ADCpt(EpicsSignal, 'TS:TSSigmaXY')
    ts_sigma_y = ADCpt(EpicsSignal, 'TS:TSSigmaY')
    ts_total = ADCpt(EpicsSignal, 'TS:TSTotal')
    total = ADCpt(EpicsSignalRO, 'Total_RBV')


class QuadEMPort(ADBase):
    port_name = Cpt(Signal, value="")

    def __init__(self, port_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.port_name.put(port_name)


class QuadEMV33(QuadEM):
    conf = Cpt(QuadEMPort, port_name="EM180")
    em_range = Cpt(SignalWithRBV, "Range", string=True)

    current1 = ADCpt(StatsPluginV33, 'Current1:')
    current2 = ADCpt(StatsPluginV33, 'Current2:')
    current3 = ADCpt(StatsPluginV33, 'Current3:')
    current4 = ADCpt(StatsPluginV33, 'Current4:')
    sum_all = ADCpt(StatsPluginV33, 'SumAll:')



from __future__ import print_function, division
import time
import time as ttime
import logging

from .utils import makedirs

from collections import OrderedDict

import h5py

from ophyd.areadetector import (DetectorBase, CamBase,
                                EpicsSignalWithRBV as SignalWithRBV)
from ophyd import (Signal, EpicsSignal, EpicsSignalRO, DerivedSignal)

from ophyd import (Device, Component as Cpt, FormattedComponent as FC,  # noqa: F401
                   DynamicDeviceComponent as DDC)
from ophyd.areadetector.plugins import PluginBase
from ophyd.areadetector.filestore_mixins import FileStorePluginBase

from ophyd.areadetector.plugins import HDF5Plugin
from ophyd.areadetector import ADBase
from ophyd.device import (BlueskyInterface, Staged)
from ophyd.status import DeviceStatus

from databroker.assets.handlers import (Xspress3HDF5Handler,
                                        XS3_XRF_DATA_KEY as XRF_DATA_KEY)


logger = logging.getLogger(__name__)


def ev_to_bin(ev):
    '''Convert eV to bin number'''
    return int(ev / 10)


def bin_to_ev(bin_):
    '''Convert bin number to eV'''
    return int(bin_) * 10


class EvSignal(DerivedSignal):
    '''A signal that converts a bin number into electron volts'''
    def __init__(self, parent_attr, *, parent=None, **kwargs):
        bin_signal = getattr(parent, parent_attr)
        super().__init__(derived_from=bin_signal, parent=parent, **kwargs)

    def get(self, **kwargs):
        bin_ = super().get(**kwargs)
        return bin_to_ev(bin_)

    def put(self, ev_value, **kwargs):
        bin_value = ev_to_bin(ev_value)
        return super().put(bin_value, **kwargs)

    def describe(self):
        desc = super().describe()
        desc[self.name]['units'] = 'eV'
        return desc


class Xspress3FileStore(FileStorePluginBase, HDF5Plugin):
    '''Xspress3 acquisition -> filestore'''
    num_capture_calc = Cpt(EpicsSignal, 'NumCapture_CALC')
    num_capture_calc_disable = Cpt(EpicsSignal, 'NumCapture_CALC.DISA')
    filestore_spec = Xspress3HDF5Handler.HANDLER_NAME

    def __init__(self, basename, *, config_time=0.5,
                 mds_key_format='{self.settings.name}_ch{chan}', parent=None,
                 **kwargs):
        super().__init__(basename, parent=parent, **kwargs)
        det = parent
        self.settings = det.settings

        # Use the EpicsSignal file_template from the detector
        self.stage_sigs[self.blocking_callbacks] = 1
        self.stage_sigs[self.enable] = 1
        self.stage_sigs[self.compression] = 'zlib'
        self.stage_sigs[self.file_template] = '%s%s_%6.6d.h5'

        self._filestore_res = None
        self.channels = list(range(1, len([_ for _ in det.component_names
                                           if _.startswith('chan')]) + 1))
        # this was in original code, but I kinda-sorta nuked because
        # it was not needed for SRX and I could not guess what it did
        self._master = None

        self._config_time = config_time
        self.mds_keys = {chan: mds_key_format.format(self=self, chan=chan)
                         for chan in self.channels}

    def stop(self, success=False):
        ret = super().stop(success=success)
        self.capture.put(0)
        return ret

    def kickoff(self):
        # TODO
        raise NotImplementedError()

    def collect(self):
        # TODO (hxn-specific implementation elsewhere)
        raise NotImplementedError()

    def make_filename(self):
        fn, rp, write_path = super().make_filename()
        if self.parent.make_directories.get():
            makedirs(write_path)
        return fn, rp, write_path

    def unstage(self):
        try:
            i = 0
            # this needs a fail-safe, RE will now hang forever here
            # as we eat all SIGINT to ensure that cleanup happens in
            # orderly manner.
            # If we are here this is a sign that we have not configured the xs3
            # correctly and it is expecting to capture more points than it
            # was triggered to take.
            while self.capture.get() == 1:
                i += 1
                if (i % 50) == 0:
                    logger.warning('Still capturing data .... waiting.')
                time.sleep(0.1)
                if i > 150:
                    logger.warning('Still capturing data .... giving up.')
                    logger.warning('Check that the xspress3 is configured to take the right '
                                   'number of frames '
                                   f'(it is trying to take {self.parent.settings.num_images.get()})')
                    self.capture.put(0)
                    break

        except KeyboardInterrupt:
            self.capture.put(0)
            logger.warning('Still capturing data .... interrupted.')

        return super().unstage()

    def generate_datum(self, key, timestamp, datum_kwargs):
        sn, n = next((f'channel{j}', j)
                     for j in self.channels
                     if getattr(self.parent, f'channel{j}').name == key)
        datum_kwargs.update({'frame': self.parent._abs_trigger_count,
                             'channel': int(sn[7:])})
        self.mds_keys[n] = key
        super().generate_datum(key, timestamp, datum_kwargs)

    def stage(self):
        # if should external trigger
        ext_trig = self.parent.external_trig.get()

        logger.debug('Stopping xspress3 acquisition')
        # really force it to stop acquiring
        self.settings.acquire.put(0, wait=True)

        total_points = self.parent.total_points.get()
        if total_points < 1:
            raise RuntimeError("You must set the total points")
        spec_per_point = self.parent.spectra_per_point.get()
        total_capture = total_points * spec_per_point

        # stop previous acquisition
        self.stage_sigs[self.settings.acquire] = 0

        # re-order the stage signals and disable the calc record which is
        # interfering with the capture count
        self.stage_sigs.pop(self.num_capture, None)
        self.stage_sigs.pop(self.settings.num_images, None)
        self.stage_sigs[self.num_capture_calc_disable] = 1

        if ext_trig:
            logger.debug('Setting up external triggering')
            self.stage_sigs[self.settings.trigger_mode] = 'TTL Veto Only'
            self.stage_sigs[self.settings.num_images] = total_capture
        else:
            logger.debug('Setting up internal triggering')
            # self.settings.trigger_mode.put('Internal')
            # self.settings.num_images.put(1)
            self.stage_sigs[self.settings.trigger_mode] = 'Internal'
            self.stage_sigs[self.settings.num_images] = spec_per_point

        self.stage_sigs[self.auto_save] = 'No'
        logger.debug('Configuring other filestore stuff')

        logger.debug('Making the filename')
        filename, read_path, write_path = self.make_filename()

        logger.debug('Setting up hdf5 plugin: ioc path: %s filename: %s',
                     write_path, filename)

        logger.debug('Erasing old spectra')
        self.settings.erase.put(1, wait=True)

        # this must be set after self.settings.num_images because at the Epics
        # layer  there is a helpful link that sets this equal to that (but
        # not the other way)
        self.stage_sigs[self.num_capture] = total_capture

        # actually apply the stage_sigs
        ret = super().stage()

        self._fn = self.file_template.get() % (self._fp,
                                               self.file_name.get(),
                                               self.file_number.get())

        if not self.file_path_exists.get():
            raise IOError("Path {} does not exits on IOC!! Please Check"
                          .format(self.file_path.get()))

        logger.debug('Inserting the filestore resource: %s', self._fn)
        self._generate_resource({})
        self._filestore_res = self._asset_docs_cache[-1][-1]

        # this gets auto turned off at the end
        self.capture.put(1)

        # Xspress3 needs a bit of time to configure itself...
        # this does not play nice with the event loop :/
        time.sleep(self._config_time)

        return ret

    def configure(self, total_points=0, master=None, external_trig=False,
                  **kwargs):
        raise NotImplementedError()

    def describe(self):
        # should this use a better value?
        size = (self.width.get(), )

        spec_desc = {'external': 'FILESTORE:',
                     'dtype': 'array',
                     'shape': size,
                     'source': 'FileStore:'
                     }

        desc = OrderedDict()
        for chan in self.channels:
            key = self.mds_keys[chan]
            desc[key] = spec_desc

        return desc


class Xspress3DetectorSettings(CamBase):
    '''Quantum Detectors Xspress3 detector'''

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = ['config_path', 'config_save_path',
                                   ]
        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs, **kwargs)

    config_path = Cpt(SignalWithRBV, 'CONFIG_PATH', string=True)
    config_save_path = Cpt(SignalWithRBV, 'CONFIG_SAVE_PATH', string=True)
    connect = Cpt(EpicsSignal, 'CONNECT')
    connected = Cpt(EpicsSignal, 'CONNECTED')
    ctrl_dtc = Cpt(SignalWithRBV, 'CTRL_DTC')
    ctrl_mca_roi = Cpt(SignalWithRBV, 'CTRL_MCA_ROI')
    debounce = Cpt(SignalWithRBV, 'DEBOUNCE')
    disconnect = Cpt(EpicsSignal, 'DISCONNECT')
    erase = Cpt(EpicsSignal, 'ERASE')
    # erase_array_counters = Cpt(EpicsSignal, 'ERASE_ArrayCounters')
    # erase_attr_reset = Cpt(EpicsSignal, 'ERASE_AttrReset')
    # erase_proc_reset_filter = Cpt(EpicsSignal, 'ERASE_PROC_ResetFilter')
    frame_count = Cpt(EpicsSignalRO, 'FRAME_COUNT_RBV')
    invert_f0 = Cpt(SignalWithRBV, 'INVERT_F0')
    invert_veto = Cpt(SignalWithRBV, 'INVERT_VETO')
    max_frames = Cpt(EpicsSignalRO, 'MAX_FRAMES_RBV')
    max_frames_driver = Cpt(EpicsSignalRO, 'MAX_FRAMES_DRIVER_RBV')
    max_num_channels = Cpt(EpicsSignalRO, 'MAX_NUM_CHANNELS_RBV')
    max_spectra = Cpt(SignalWithRBV, 'MAX_SPECTRA')
    xsp_name = Cpt(EpicsSignal, 'NAME')
    num_cards = Cpt(EpicsSignalRO, 'NUM_CARDS_RBV')
    num_channels = Cpt(SignalWithRBV, 'NUM_CHANNELS')
    num_frames_config = Cpt(SignalWithRBV, 'NUM_FRAMES_CONFIG')
    reset = Cpt(EpicsSignal, 'RESET')
    restore_settings = Cpt(EpicsSignal, 'RESTORE_SETTINGS')
    run_flags = Cpt(SignalWithRBV, 'RUN_FLAGS')
    save_settings = Cpt(EpicsSignal, 'SAVE_SETTINGS')
    trigger_signal = Cpt(EpicsSignal, 'TRIGGER')
    # update = Cpt(EpicsSignal, 'UPDATE')
    # update_attr = Cpt(EpicsSignal, 'UPDATE_AttrUpdate')


class Xspress3ROISettings(PluginBase):
    '''Full areaDetector plugin settings'''
    array_data = Cpt(EpicsSignalRO, 'ArrayData_RBV')

    @property
    def ad_root(self):
        root = self.parent
        while True:
            if not isinstance(root.parent, ADBase):
                return root
            root = root.parent


class Xspress3ROI(ADBase):
    '''A configurable Xspress3 EPICS ROI'''

    # prefix: C{channel}_   MCA_ROI{self.roi_num}
    bin_low = FC(SignalWithRBV, '{self.channel.prefix}{self.bin_suffix}_LLM')
    bin_high = FC(SignalWithRBV, '{self.channel.prefix}{self.bin_suffix}_HLM')

    # derived from the bin signals, low and high electron volt settings:
    ev_low = Cpt(EvSignal, parent_attr='bin_low')
    ev_high = Cpt(EvSignal, parent_attr='bin_high')

    # C{channel}_  ROI{self.roi_num}
    value = Cpt(EpicsSignalRO, 'Value_RBV')
    value_sum = Cpt(EpicsSignalRO, 'ValueSum_RBV')

    enable = Cpt(SignalWithRBV, 'EnableCallbacks')
    # ad_plugin = Cpt(Xspress3ROISettings, '')

    @property
    def ad_root(self):
        root = self.parent
        while True:
            if not isinstance(root.parent, ADBase):
                return root
            root = root.parent

    def __init__(self, prefix, *, roi_num=0, use_sum=False,
                 read_attrs=None, configuration_attrs=None, parent=None,
                 bin_suffix=None, **kwargs):

        if read_attrs is None:
            if use_sum:
                read_attrs = ['value_sum']
            else:
                read_attrs = ['value', 'value_sum']

        if configuration_attrs is None:
            configuration_attrs = ['ev_low', 'ev_high', 'enable']

        rois = parent
        channel = rois.parent
        self._channel = channel
        self._roi_num = roi_num
        self._use_sum = use_sum
        self._ad_plugin = getattr(rois, 'ad_attr{:02d}'.format(roi_num))

        if bin_suffix is None:
            bin_suffix = 'MCA_ROI{}'.format(roi_num)

        self.bin_suffix = bin_suffix

        super().__init__(prefix, parent=parent, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs, **kwargs)

    @property
    def settings(self):
        '''Full areaDetector settings'''
        return self._ad_plugin

    @property
    def channel(self):
        '''The Xspress3Channel instance associated with the ROI'''
        return self._channel

    @property
    def channel_num(self):
        '''The channel number associated with the ROI'''
        return self._channel.channel_num

    @property
    def roi_num(self):
        '''The ROI number'''
        return self._roi_num

    def clear(self):
        '''Clear and disable this ROI'''
        self.configure(0, 0)

    def configure(self, ev_low, ev_high):
        '''Configure the ROI with low and high eV

        Parameters
        ----------
        ev_low : int
            low electron volts for ROI
        ev_high : int
            high electron volts for ROI
        '''
        ev_low = int(ev_low)
        ev_high = int(ev_high)

        enable = 1 if ev_high > ev_low else 0
        changed = any([self.ev_high.get() != ev_high,
                       self.ev_low.get() != ev_low,
                       self.enable.get() != enable])

        if not changed:
            return

        logger.debug('Setting up EPICS ROI: name=%s ev=(%s, %s) '
                     'enable=%s prefix=%s channel=%s',
                     self.name, ev_low, ev_high, enable, self.prefix,
                     self._channel)
        if ev_high <= self.ev_low.get():
            self.ev_low.put(0)

        self.ev_high.put(ev_high)
        self.ev_low.put(ev_low)
        self.enable.put(enable)


def make_rois(rois):
    defn = OrderedDict()
    for roi in rois:
        attr = 'roi{:02d}'.format(roi)
        #             cls          prefix                kwargs
        defn[attr] = (Xspress3ROI, 'ROI{}:'.format(roi), dict(roi_num=roi))
        # e.g., device.rois.roi01 = Xspress3ROI('ROI1:', roi_num=1)

        # AreaDetector NDPluginAttribute information
        attr = 'ad_attr{:02d}'.format(roi)
        defn[attr] = (Xspress3ROISettings, 'ROI{}:'.format(roi),
                      dict(read_attrs=[]))
        # e.g., device.rois.roi01 = Xspress3ROI('ROI1:', roi_num=1)

        # TODO: 'roi01' and 'ad_attr_01' have the same prefix and could
        # technically be combined. Is this desirable?

    defn['num_rois'] = (Signal, None, dict(value=len(rois)))
    # e.g., device.rois.num_rois.get() => 16
    return defn


class Xspress3Channel(ADBase):
    roi_name_format = 'Det{self.channel_num}_{roi_name}'
    roi_sum_name_format = 'Det{self.channel_num}_{roi_name}_sum'

    rois = DDC(make_rois(range(1, 17)))
    vis_enabled = Cpt(EpicsSignal, 'PluginControlVal')
    extra_rois_enabled = Cpt(EpicsSignal, 'PluginControlValExtraROI')
    
    def __init__(self, prefix, *, channel_num=None, **kwargs):
        self.channel_num = int(channel_num)

        super().__init__(prefix, **kwargs)

    @property
    def all_rois(self):
        for roi in range(1, self.rois.num_rois.get() + 1):
            yield getattr(self.rois, 'roi{:02d}'.format(roi))

    def set_roi(self, index, ev_low, ev_high, *, name=None):
        '''Set specified ROI to (ev_low, ev_high)

        Parameters
        ----------
        index : int or Xspress3ROI
            The roi index or instance to set
        ev_low : int
            low eV setting
        ev_high : int
            high eV setting
        name : str, optional
            The unformatted ROI name to set. Each channel specifies its own
            `roi_name_format` and `roi_sum_name_format` in which the name
            parameter will get expanded.
        '''
        if isinstance(index, Xspress3ROI):
            roi = index
        else:
            if index <= 0:
                raise ValueError('ROI index starts from 1')
            roi = list(self.all_rois)[index - 1]

        roi.configure(ev_low, ev_high)
        if name is not None:
            roi_name = self.roi_name_format.format(self=self, roi_name=name)
            roi.name = roi_name
            roi.value.name = roi_name
            roi.value_sum.name = self.roi_sum_name_format.format(self=self,
                                                                 roi_name=name)

    def clear_all_rois(self):
        '''Clear all ROIs'''
        for roi in self.all_rois:
            roi.clear()


class Xspress3Detector(DetectorBase):
    settings = Cpt(Xspress3DetectorSettings, '')

    external_trig = Cpt(Signal, value=False,
                        doc='Use external triggering')
    total_points = Cpt(Signal, value=-1,
                       doc='The total number of points to acquire overall')
    spectra_per_point = Cpt(Signal, value=1,
                            doc='Number of spectra per point')
    make_directories = Cpt(Signal, value=False,
                           doc='Make directories on the DAQ side')
    rewindable = Cpt(Signal, value=False,
                     doc='Xspress3 cannot safely be rewound in bluesky')

    # XF:03IDC-ES{Xsp:1}           C1_   ...
    # channel1 = Cpt(Xspress3Channel, 'C1_', channel_num=1)

    data_key = XRF_DATA_KEY

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 name=None, parent=None,
                 # to remove?
                 file_path='', ioc_file_path='', default_channels=None,
                 channel_prefix=None,
                 roi_sums=False,
                 # to remove?
                 **kwargs):

        if read_attrs is None:
            read_attrs = ['channel1', ]

        if configuration_attrs is None:
            configuration_attrs = ['channel1.rois', 'settings']

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         name=name, parent=parent, **kwargs)

        # get all sub-device instances
        sub_devices = {attr: getattr(self, attr)
                       for attr in self._sub_devices}

        # filter those sub-devices, just giving channels
        channels = {dev.channel_num: dev
                    for attr, dev in sub_devices.items()
                    if isinstance(dev, Xspress3Channel)
                    }

        # make an ordered dictionary with the channels in order
        self._channels = OrderedDict(sorted(channels.items()))

    @property
    def channels(self):
        return self._channels.copy()

    @property
    def all_rois(self):
        for ch_num, channel in self._channels.items():
            for roi in channel.all_rois:
                yield roi

    @property
    def enabled_rois(self):
        for roi in self.all_rois:
            if roi.enable.get():
                yield roi

    def read_hdf5(self, fn, *, rois=None, max_retries=2):
        '''Read ROI data from an HDF5 file using the current ROI configuration

        Parameters
        ----------
        fn : str
            HDF5 filename to load
        rois : sequence of Xspress3ROI instances, optional

        '''
        if rois is None:
            rois = self.enabled_rois

        num_points = self.settings.num_images.get()
        if isinstance(fn, h5py.File):
            hdf = fn
        else:
            hdf = h5py.File(fn, 'r')

        RoiTuple = Xspress3ROI.get_device_tuple()

        handler = Xspress3HDF5Handler(hdf, key=self.data_key)
        for roi in self.enabled_rois:
            roi_data = handler.get_roi(chan=roi.channel_num,
                                       bin_low=roi.bin_low.get(),
                                       bin_high=roi.bin_high.get(),
                                       max_points=num_points)

            roi_info = RoiTuple(bin_low=roi.bin_low.get(),
                                bin_high=roi.bin_high.get(),
                                ev_low=roi.ev_low.get(),
                                ev_high=roi.ev_high.get(),
                                value=roi_data,
                                value_sum=None,
                                enable=None)

            yield roi.name, roi_info


class XspressTrigger(BlueskyInterface):
    """Base class for trigger mixin classes

    Subclasses must define a method with this signature:

    `acquire_changed(self, value=None, old_value=None, **kwargs)`
    """
    # TODO **
    # count_time = self.settings.acquire_period

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # settings
        self._status = None
        self._acquisition_signal = self.settings.acquire
        self._abs_trigger_count = 0

    def stage(self):
        self._abs_trigger_count = 0
        self._acquisition_signal.subscribe(self._acquire_changed)
        return super().stage()

    def unstage(self):
        ret = super().unstage()
        self._acquisition_signal.clear_sub(self._acquire_changed)
        self._status = None
        return ret

    def _acquire_changed(self, value=None, old_value=None, **kwargs):
        "This is called when the 'acquire' signal changes."
        if self._status is None:
            return
        if (old_value == 1) and (value == 0):
            # Negative-going edge means an acquisition just finished.
            self._status._finished()

    def trigger(self):
        if self._staged != Staged.yes:
            raise RuntimeError("not staged")

        self._status = DeviceStatus(self)
        self._acquisition_signal.put(1, wait=False)
        trigger_time = ttime.time()

        for sn in self.read_attrs:
            if sn.startswith('channel') and '.' not in sn:
                ch = getattr(self, sn)
                self.dispatch(ch.name, trigger_time)

        self._abs_trigger_count += 1
        return self._status



from enum import IntEnum
import logging

from ophyd import (Device, Component as Cpt, FormattedComponent as FC,
                   Signal)
from ophyd import (EpicsSignal, EpicsSignalRO, DeviceStatus)
from ophyd.utils import set_and_wait

from .trigger_mixins import ModalBase

logger = logging.getLogger(__name__)


def _get_configuration_attrs(inst, *, signal_class=Signal):
    cls = inst.__class__
    return [sig_name for sig_name in cls.component_names
            if issubclass(getattr(cls, sig_name).cls, signal_class)]


class ZebraInputEdge(IntEnum):
    FALLING = 1
    RISING = 0


class ZebraAddresses(IntEnum):
    DISCONNECT = 0
    IN1_TTL = 1
    IN1_NIM = 2
    IN1_LVDS = 3
    IN2_TTL = 4
    IN2_NIM = 5
    IN2_LVDS = 6
    IN3_TTL = 7
    IN3_OC = 8
    IN3_LVDS = 9
    IN4_TTL = 10
    IN4_CMP = 11
    IN4_PECL = 12
    IN5_ENCA = 13
    IN5_ENCB = 14
    IN5_ENCZ = 15
    IN5_CONN = 16
    IN6_ENCA = 17
    IN6_ENCB = 18
    IN6_ENCZ = 19
    IN6_CONN = 20
    IN7_ENCA = 21
    IN7_ENCB = 22
    IN7_ENCZ = 23
    IN7_CONN = 24
    IN8_ENCA = 25
    IN8_ENCB = 26
    IN8_ENCZ = 27
    IN8_CONN = 28
    PC_ARM = 29
    PC_GATE = 30
    PC_PULSE = 31
    AND1 = 32
    AND2 = 33
    AND3 = 34
    AND4 = 35
    OR1 = 36
    OR2 = 37
    OR3 = 38
    OR4 = 39
    GATE1 = 40
    GATE2 = 41
    GATE3 = 42
    GATE4 = 43
    DIV1_OUTD = 44
    DIV2_OUTD = 45
    DIV3_OUTD = 46
    DIV4_OUTD = 47
    DIV1_OUTN = 48
    DIV2_OUTN = 49
    DIV3_OUTN = 50
    DIV4_OUTN = 51
    PULSE1 = 52
    PULSE2 = 53
    PULSE3 = 54
    PULSE4 = 55
    QUAD_OUTA = 56
    QUAD_OUTB = 57
    CLOCK_1KHZ = 58
    CLOCK_1MHZ = 59
    SOFT_IN1 = 60
    SOFT_IN2 = 61
    SOFT_IN3 = 62
    SOFT_IN4 = 63


class EpicsSignalWithRBV(EpicsSignal):
    # An EPICS signal that uses the Zebra convention of 'pvname' being the
    # setpoint and 'pvname:RBV' being the read-back

    def __init__(self, prefix, **kwargs):
        super().__init__(prefix + ':RBV', write_pv=prefix, **kwargs)


class ZebraPulse(Device):
    width = Cpt(EpicsSignalWithRBV, 'WID')
    input_addr = Cpt(EpicsSignalWithRBV, 'INP')
    input_str = Cpt(EpicsSignalRO, 'INP:STR', string=True)
    input_status = Cpt(EpicsSignalRO, 'INP:STA')
    delay = Cpt(EpicsSignalWithRBV, 'DLY')
    delay_sync = Cpt(EpicsSignal, 'DLY:SYNC')
    time_units = Cpt(EpicsSignalWithRBV, 'PRE', string=True)
    output = Cpt(EpicsSignal, 'OUT')

    input_edge = FC(EpicsSignal,
                    '{self._zebra_prefix}POLARITY:{self._edge_addr}')

    _edge_addrs = {1: 'BC',
                   2: 'BD',
                   3: 'BE',
                   4: 'BF',
                   }

    def __init__(self, prefix, *, index=None, parent=None,
                 configuration_attrs=None, read_attrs=None, **kwargs):
        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = _get_configuration_attrs(self)

        zebra = parent
        self.index = index
        self._zebra_prefix = zebra.prefix
        self._edge_addr = self._edge_addrs[index]

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, parent=parent, **kwargs)


class ZebraOutputBase(Device):
    '''The base of all zebra outputs (1~8)

        Front outputs
        # TTL  LVDS  NIM  PECL  OC  ENC
        1  o    o     o
        2  o    o     o
        3  o    o               o
        4  o          o    o

        Rear outputs
        # TTL  LVDS  NIM  PECL  OC  ENC
        5                            o
        6                            o
        7                            o
        8                            o

    '''
    def __init__(self, prefix, *, index=None, read_attrs=None,
                 configuration_attrs=None, **kwargs):
        self.index = index

        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = _get_configuration_attrs(self)

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs, **kwargs)


class ZebraOutputType(Device):
    '''Shared by all output types (ttl, lvds, nim, pecl, out)'''
    addr = Cpt(EpicsSignalWithRBV, '')
    status = Cpt(EpicsSignalRO, ':STA')
    string = Cpt(EpicsSignalRO, ':STR', string=True)
    sync = Cpt(EpicsSignal, ':SYNC')
    write_output = Cpt(EpicsSignal, ':SET')

    def __init__(self, prefix, *, read_attrs=None, configuration_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = _get_configuration_attrs(self)

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs, **kwargs)


class ZebraFrontOutput12(ZebraOutputBase):
    ttl = Cpt(ZebraOutputType, 'TTL')
    lvds = Cpt(ZebraOutputType, 'LVDS')
    nim = Cpt(ZebraOutputType, 'NIM')


class ZebraFrontOutput3(ZebraOutputBase):
    ttl = Cpt(ZebraOutputType, 'TTL')
    lvds = Cpt(ZebraOutputType, 'LVDS')
    open_collector = Cpt(ZebraOutputType, 'OC')


class ZebraFrontOutput4(ZebraOutputBase):
    ttl = Cpt(ZebraOutputType, 'TTL')
    nim = Cpt(ZebraOutputType, 'NIM')
    pecl = Cpt(ZebraOutputType, 'PECL')


class ZebraRearOutput(ZebraOutputBase):
    enca = Cpt(ZebraOutputType, 'ENCA')
    encb = Cpt(ZebraOutputType, 'ENCB')
    encz = Cpt(ZebraOutputType, 'ENCZ')
    conn = Cpt(ZebraOutputType, 'CONN')


class ZebraGateInput(Device):
    addr = Cpt(EpicsSignalWithRBV, '')
    string = Cpt(EpicsSignalRO, ':STR', string=True)
    status = Cpt(EpicsSignalRO, ':STA')
    sync = Cpt(EpicsSignal, ':SYNC')
    write_input = Cpt(EpicsSignal, ':SET')

    # Input edge index depends on the gate number (these are set in __init__)
    edge = FC(EpicsSignal,
              '{self._zebra_prefix}POLARITY:B{self._input_edge_idx}')

    def __init__(self, prefix, *, index=None, parent=None,
                 configuration_attrs=None, read_attrs=None, **kwargs):
        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = _get_configuration_attrs(self)

        gate = parent
        zebra = gate.parent

        self.index = index
        self._zebra_prefix = zebra.prefix
        self._input_edge_idx = gate._input_edge_idx[self.index]

        super().__init__(prefix, read_attrs=read_attrs,
                         configuration_attrs=configuration_attrs,
                         parent=parent, **kwargs)


class ZebraGate(Device):
    input1 = Cpt(ZebraGateInput, 'INP1', index=1)
    input2 = Cpt(ZebraGateInput, 'INP2', index=2)
    output = Cpt(EpicsSignal, 'OUT')

    def __init__(self, prefix, *, index=None, read_attrs=None,
                 configuration_attrs=None, **kwargs):
        self.index = index
        self._input_edge_idx = {1: index - 1,
                                2: 4 + index - 1
                                }

        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = ['output']

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

    def set_input_edges(self, edge1, edge2):
        set_and_wait(self.input1.edge, int(edge1))
        set_and_wait(self.input2.edge, int(edge2))


class Zebra(ModalBase, Device):
    soft_input1 = Cpt(EpicsSignal, 'SOFT_IN:B0')
    soft_input2 = Cpt(EpicsSignal, 'SOFT_IN:B1')
    soft_input3 = Cpt(EpicsSignal, 'SOFT_IN:B2')
    soft_input4 = Cpt(EpicsSignal, 'SOFT_IN:B3')

    pulse1 = Cpt(ZebraPulse, 'PULSE1_', index=1)
    pulse2 = Cpt(ZebraPulse, 'PULSE2_', index=2)
    pulse3 = Cpt(ZebraPulse, 'PULSE3_', index=3)
    pulse4 = Cpt(ZebraPulse, 'PULSE4_', index=4)

    output1 = Cpt(ZebraFrontOutput12, 'OUT1_', index=1)
    output2 = Cpt(ZebraFrontOutput12, 'OUT2_', index=2)
    output3 = Cpt(ZebraFrontOutput3, 'OUT3_', index=3)
    output4 = Cpt(ZebraFrontOutput4, 'OUT4_', index=4)

    output5 = Cpt(ZebraRearOutput, 'OUT5_', index=5)
    output6 = Cpt(ZebraRearOutput, 'OUT6_', index=6)
    output7 = Cpt(ZebraRearOutput, 'OUT7_', index=7)
    output8 = Cpt(ZebraRearOutput, 'OUT8_', index=8)

    gate1 = Cpt(ZebraGate, 'GATE1_', index=1)
    gate2 = Cpt(ZebraGate, 'GATE2_', index=2)
    gate3 = Cpt(ZebraGate, 'GATE3_', index=3)
    gate4 = Cpt(ZebraGate, 'GATE4_', index=4)

    addresses = ZebraAddresses

    def __init__(self, prefix, *, configuration_attrs=None, read_attrs=None,
                 **kwargs):
        if read_attrs is None:
            read_attrs = []
        if configuration_attrs is None:
            configuration_attrs = _get_configuration_attrs(self)

        super().__init__(prefix, configuration_attrs=configuration_attrs,
                         read_attrs=read_attrs, **kwargs)

        self.pulse = dict(self._get_indexed_devices(ZebraPulse))
        self.output = dict(self._get_indexed_devices(ZebraOutputBase))
        self.gate = dict(self._get_indexed_devices(ZebraGate))

    def _get_indexed_devices(self, cls):
        for attr in self._sub_devices:
            dev = getattr(self, attr)
            if isinstance(dev, cls):
                yield dev.index, dev

    def mode_internal(self):
        super().mode_internal()
        # handle the scan type here

    def mode_external(self):
        super().mode_external()
        # handle the scan type here

    def trigger(self):
        # Re-implement this to trigger as desired in bluesky
        status = DeviceStatus(self)
        status._finished()
        return status
